from collections import Counter

import nltk
from sklearn.metrics import cohen_kappa_score


def partial_match(b1, e1, b2, e2):
    """
    Check if the interval (b1,e1) intersects
    with the interval (b2, e2)

    @param int: beginning of a interval
    @param int: ending of a interval
    @param int: beginning of a interval
    @param int: ending of a interval

    @return bool: if there is an intersection between an interval
     it returns true, otherwise returns false
    """

    if (b2 <= e1) and (b2 >= b1):
        return True
    if (e2 <= e1) and (e2 >= b1):
        return True

    if (b1 >= b2) and (e1 <= e2):
        return True
    return False


def partial_search_annotation(ann, ann_lst):
    """
    given  an annotation (dictionary), do a binary search in a list of annotations
    if the tokens are in the annotation list. The search looks for only
    if there is in intersection between ann and any of the intervals in
    ann_lst

    @param (int, int): a tuple of integers that indicates an intervals
    @param [(int,int, int)]: a list of intervals

    @return int: return a integer that is the index position of the element, or -1
    if the interval is absence of ann_lst

    """

    ans = -1
    # ans = len(word_tokenize(ann["value"]))
    start, end = ann["offset1"]
    b = 0
    e = len(ann_lst) - 1
    m = int((b + e) / 2)
    idkey = None

    while (b <= e):
        start_search, end_search, idkey = ann_lst[m]

        if end < start_search:
            e = m - 1
            m = int((b + e) / 2)
        else:
            if start > end_search:
                b = m + 1
                m = int((b + e) / 2)
            else:
                if partial_match(start, end, start_search, end_search):
                    ans = m
                    break

    return ans, idkey


def search_annotation(ann, ann_lst):
    """
    given  an annotation (dictionary), do a binary search in a list of annotations
    if the tokens are in the annotation list

    @param (int, int): a tuple of integers that indicates an intervals
    @param [(int,int, int)]: a list of intervals

    @return int: return a integer that is the index position of the element, or -1
    if the interval is absence of ann_lst
    """

    ans = -1
    # ans = len(word_tokenize(ann["value"]))
    start, end = ann["offset1"]
    b = 0
    e = len(ann_lst) - 1
    m = int((b + e) / 2)
    idkey = None

    while (b <= e):
        start_search, end_search, idkey = ann_lst[m]
        if end < start_search:
            e = m - 1
            m = int((b + e) / 2)
        else:
            if start > end_search:
                b = m + 1
                m = int((b + e) / 2)
            else:
                if start == start_search and end == end_search:
                    ans = m
                break

    return ans, idkey


def get_intervals(ann_lst):
    """
    get the intervals of annotations as a sorted tuple list

    @param dictionary: dictionary of annotations

    @return tuple list of integers
    """

    interval_lst = []

    for el in ann_lst:
        s1, e1 = el["offset1"]
        interval_lst.append((s1, e1, el["id"]))
        if "offset2" in el:
            s2, e2 = el["offset2"]
            interval_lst.append((s2, e2, el["id"]))

    sorted(interval_lst, key=lambda elem: elem[0])

    return interval_lst

def get_map_list(ann_link):

    arg_map = dict()
    for ann in ann_link:
        arg_map[ann["args"]] = ann["id"]

    return arg_map

def matrix_confusion(res_pred, res_target):


    tp = 0
    fp = 0
    for ans in res_pred:
        if res_pred[ans] is not None:
            # ann_target
            tp += 1  # true positive
        else:
            # false positive
            fp += 1

    tn = 0
    fn = 0
    for ans in res_target:
        if res_target[ans] is None:
            fn += 1
        else:
            tn += 1

    return tp, tn, fp, fn

def compute_precision(tp, fp):

    if tp == 0 and fp == 0:
        return 0
    else:
        return tp / (tp + fp)

def compute_recall(tp, fn):
    if tp == 0 and fn == 0:
        return 0
    else:
        return tp / (tp + fn)

def compute_f1(res_pred, res_target):

    tp, tn, fp, fn = matrix_confusion(res_pred, res_target)
    precision = compute_precision(tp, fp)
    recall = compute_recall(tp, fn)

    if precision == 0 and recall == 0:
        f1 = 0
    else:
        f1 = 2 * (precision * recall) / (precision + recall)

    return precision, recall, f1

def compute_relax_scores(elem, ann_pred, ann_target):
    if elem == "participant" or elem == "event" or elem == "time":
        match_pred, match_target = compute_relax_scores_entity(ann_pred, ann_target)
        return compute_f1(match_pred, match_target)
    else:
        match_pred, match_target = compute_relax_scores_srlink(ann_pred, ann_target)
        return compute_f1(match_pred, match_target)

def compute_relax_scores_entity(ann_pred, ann_target):
    """
    it computes the  relaxed scores (tokens in common is a match)
    for two annotations

    @param dictionary: annotations of the prediction
    @param dictionary: annotations of the target/human-labeled

    @return match_pred, match_target
    """

    interval_pred_lst = get_intervals(ann_pred)
    interval_target_lst = get_intervals(ann_target)

    interval_pred_lst.sort(key=lambda x: int(x[0]))
    interval_target_lst.sort(key=lambda x: int(x[0]))

    search_pred = {}
    for pred in ann_pred:
        ans, id_ans = partial_search_annotation(pred, interval_target_lst)
        if ans == -1:
            search_pred[pred["id"]] =  None
        else:
            search_pred[pred["id"]] =  id_ans


    fn = 0  # false negative
    search_target = {}
    for target in ann_target:
        ans, id_ans = partial_search_annotation(target, interval_pred_lst)
        if ans == -1:
            search_target[target["id"]] = None
        else:
            search_target[target["id"]] = id_ans

    return search_pred, search_target


def compute_relax_scores_srlink(ann_pred, ann_target):
    #srlink_pred, srlink_target = get_element("srlink",ann_pred, ann_target)
    #participant_pred, participant_target = get_element("participant", ann_pred, ann_target)
    #event_pred, event_target = get_element("event", ann_pred, ann_target)

    participant_pred, event_pred, srlink_pred = ann_pred
    participant_target, event_target, srlink_target = ann_target

    # preciso do indice dos atores e eventos, assim posso conferir
    # o srlink
    search_part_pred, search_part_target = compute_relax_scores_entity(participant_pred, participant_target)

    search_event_pred, search_event_target = compute_relax_scores_entity(event_pred, event_target)

    arg_map_target = get_map_list(srlink_target)
    search_pred = {}
    for srlink in srlink_pred:

        e1, e2 = srlink["args"]
        id_srlink = srlink["id"]

        # if element 1 is a partial match(id) or not (None)
        ans_e1 = None
        if e1 in search_part_pred:
            ans_e1 = search_part_pred[e1]
        else:
            if e1 in search_event_pred:
                ans_e1 = search_event_pred[e1]

        ans_e2 = None
        if e2 in search_part_pred:
            ans_e2 = search_part_pred[e2]
        else:
            if e2 in search_event_pred:
                ans_e2 = search_event_pred[e2]

        # if both match partially in the target, and
        # there is really a link between them in the 
        # srlink_target.. Then get the mapping between the 
        # event/participants from prediction to target
        # check if there is a link between them in the target 
        if ans_e1 is not None and ans_e2 is not None:
             # ans_e1 and ans_e2 are the id of the elements in the 
             # target file
             if (ans_e1, ans_e2) in arg_map_target:
                 search_pred[id_srlink] =  arg_map_target[(ans_e1, ans_e2)]
             elif (ans_e2, ans_e1) in arg_map_target:
                 search_pred[id_srlink] =  arg_map_target[(ans_e2, ans_e1)]
             else:
                 search_pred[id_srlink] = None
        else:
            search_pred[id_srlink] = None

    arg_map_pred = get_map_list(srlink_pred)
    search_target = {}
    for srlink in srlink_target:

        e1, e2 = srlink["args"]
        id_srlink = srlink["id"]

        # if element 1 is a partial match(1) or not (-1)
        ans_e1 = -1
        if e1 in search_part_target:
            ans_e1 = search_part_target[e1]
        else:
            if e1 in search_event_target:
                ans_e1 = search_event_target[e1]

        ans_e2 = -1
        if e2 in search_part_target:
            ans_e2 = search_part_target[e2]
        else:
            if e2 in search_event_pred:
                ans_e2 = search_event_target[e2]

        # if both match partially in the target, and
        # there is really a link between them in the 
        # srlink_target.. Then get the mapping between the 
        # event/participants from prediction to target
        # check if there is a link between them in the target 
        if ans_e1 is not None and ans_e2 is not None:
             # ans_e1 and ans_e2 are the id of the elements in the 
             # target file
             if (ans_e1, ans_e2) in arg_map_pred:
                 search_target[id_srlink] =  arg_map_pred[(ans_e1, ans_e2)]
             elif (ans_e2, ans_e1) in arg_map_pred:
                 search_target[id_srlink] =  arg_map_pred[(ans_e2, ans_e1)]
             else:
                 search_target[id_srlink] = None
        else:
            search_target[id_srlink] = None

    return search_pred, search_target


def compute_strict_scores(elem, ann_pred, ann_target):

    if elem == "participant" or elem == "event" or elem == "time":
        match_pred, match_target = compute_strict_scores_entity(ann_pred, ann_target)
        return compute_f1(match_pred, match_target)
    else:
        match_pred, match_target = compute_strict_scores_srlink(ann_pred, ann_target)
        return compute_f1(match_pred, match_target)

def compute_strict_scores_entity(ann_pred, ann_target):
    """
    it computes the strict scores for two annotations

    @param dictionary: annotations of the prediction
    @param dictionary: annotations of the target/human-labeled

    @return dictionary: scores (precision, recall, and f1)
    computed in a strict manner
    """

    interval_pred_lst = get_intervals(ann_pred)
    interval_target_lst = get_intervals(ann_target)

    interval_pred_lst.sort(key=lambda x: int(x[0]))
    interval_target_lst.sort(key=lambda x: int(x[0]))


    search_pred = {}
    for pred in ann_pred:
        ans, idkey = search_annotation(pred, interval_target_lst)
        if ans == -1:
            search_pred[pred["id"]] = None
        else:
            search_pred[pred["id"]] = idkey

    search_target = {}
    for target in ann_target:
        ans, idkey = search_annotation(target, interval_pred_lst)
        if ans == -1:
            search_target[target["id"]] = None
        else:
            search_target[target["id"]] = idkey

    return search_pred, search_target

def compute_strict_scores_srlink(ann_pred, ann_target):

    participant_pred, event_pred, srlink_pred = ann_pred
    participant_target, event_target, srlink_target = ann_target

    # preciso do indice dos atores e eventos, assim posso conferir
    # o srlink
    search_part_pred, search_part_target = compute_strict_scores_entity(participant_pred, participant_target)

    search_event_pred, search_event_target = compute_strict_scores_entity(event_pred, event_target)

    arg_map_target = get_map_list(srlink_target)
    search_pred = {}
    for srlink in srlink_pred:

        e1, e2 = srlink["args"]
        id_srlink = srlink["id"]

        # if element 1 is a strict match(id) or not (None)
        ans_e1 = None
        if e1 in search_part_pred:
            ans_e1 = search_part_pred[e1]
        else:
            if e1 in search_event_pred:
                ans_e1 = search_event_pred[e1]

        ans_e2 = None
        if e2 in search_part_pred:
            ans_e2 = search_part_pred[e2]
        else:
            if e2 in search_event_pred:
                ans_e2 = search_event_pred[e2]

        # if both match partially in the target, and
        # there is really a link between them in the 
        # srlink_target.. Then get the mapping between the 
        # event/participants from prediction to target
        # check if there is a link between them in the target 
        if ans_e1 is not None and ans_e2 is not None:
             # ans_e1 and ans_e2 are the id of the elements in the 
             # target file
             if (ans_e1, ans_e2) in arg_map_target:
                 search_pred[id_srlink] =  arg_map_target[(ans_e1, ans_e2)]
             elif (ans_e2, ans_e1) in arg_map_target:
                 search_pred[id_srlink] =  arg_map_target[(ans_e2, ans_e1)]
             else:
                 search_pred[id_srlink] = None
        else:
            search_pred[id_srlink] = None

    arg_map_pred = get_map_list(srlink_pred)
    search_target = {}
    for srlink in srlink_target:

        e1, e2 = srlink["args"]
        id_srlink = srlink["id"]

        # if element 1 is a partial match(1) or not (-1)
        ans_e1 = -1
        if e1 in search_part_target:
            ans_e1 = search_part_target[e1]
        else:
            if e1 in search_event_target:
                ans_e1 = search_event_target[e1]

        ans_e2 = -1
        if e2 in search_part_target:
            ans_e2 = search_part_target[e2]
        else:
            if e2 in search_event_pred:
                ans_e2 = search_event_target[e2]

        # if both match partially in the target, and
        # there is really a link between them in the 
        # srlink_target.. Then get the mapping between the 
        # event/participants from prediction to target
        # check if there is a link between them in the target 
        if ans_e1 is not None and ans_e2 is not None:
             # ans_e1 and ans_e2 are the id of the elements in the 
             # target file
             if (ans_e1, ans_e2) in arg_map_pred:
                 search_target[id_srlink] =  arg_map_pred[(ans_e1, ans_e2)]
             elif (ans_e2, ans_e1) in arg_map_pred:
                 search_target[id_srlink] =  arg_map_pred[(ans_e2, ans_e1)]
             else:
                 search_target[id_srlink] = None
        else:
            search_target[id_srlink] = None

    return search_pred, search_target

def compute_bleu_lines(data_lines1, data_lines2):

    text_span_map = {}

    for d1 in data_lines1:
        t1 = nltk.word_tokenize(d1['text_span'])
        span_id1 = d1['span_id']

        for d2 in data_lines2:

            t2 = nltk.word_tokenize(d2['text_span'])

            bleu_score = nltk.translate.bleu([t1], t2, (1,))
            if span_id1 in text_span_map:
                maxb, span_id2 = text_span_map[span_id1]
                if maxb < bleu_score:
                    text_span_map[span_id1] = (bleu_score, d2['span_id'])
            else:
                 text_span_map[span_id1] = (bleu_score, d2['span_id'])

    for span_id1 in text_span_map:
        bleu, _ = text_span_map[span_id1]
        if bleu < 0.8:
            text_span_map[span_id1] = (0, None)
    return text_span_map

def compute_iaa_link(match_pred, reader1, reader2, link_type):

    all_entities_pairs = []
    for element_id_a in match_pred:
        for element_id_b in match_pred:
            all_entities_pairs.append((element_id_a, element_id_b))

    links_lst = []
    links_type_lst = []
    for (element_id_a, element_id_b) in all_entities_pairs:
        # check if annotator1 annotated some relation between event_a and event_b
        relation1 = reader1.get_relation(element_id_a, element_id_b)
        element_id_a_2 = match_pred[element_id_a]
        element_id_b_2 = match_pred[element_id_b]

        relation2 = reader2.get_relation(element_id_a_2, element_id_b_2)

        if relation1 == None and relation2 == None:
            links_lst.append(("none", "none"))
            links_type_lst.append(("none", "none"))
        else:
            if relation1 == None:
                if relation2.startswith(link_type):
                    links_lst.append(("none", link_type.lower()))
                    links_type_lst.append(("none", relation2))
                else:
                    links_lst.append(("none", "other"))
                    links_type_lst.append(("none", "other"))
            else:
                if relation2 == None:
                    if relation1.startswith(link_type):
                        links_lst.append((link_type.lower(), "none"))
                        links_type_lst.append((relation1, "none"))
                    else:
                        links_lst.append(("other", "none"))
                        links_type_lst.append(("other", "none"))
                else:
                    if relation1.startswith(link_type) and relation2.startswith(link_type):
                        links_lst.append((link_type.lower(), link_type.lower()))
                        links_type_lst.append((relation1, relation2))
                    else:
                        links_lst.append(("other", "other"))
                        links_type_lst.append(("other", "other"))

    link_annotations1 = [link[0] for link in links_lst]
    link_annotations2 = [link[1] for link in links_lst]

    link_kappa = cohen_kappa_score(link_annotations1, link_annotations2)

    link_type_annotations1 = [link[0] for link in links_type_lst]
    link_type_annotations2 = [link[1] for link in links_type_lst]

    link_type_kappa = cohen_kappa_score(link_type_annotations1, link_type_annotations2)

    return link_kappa, link_type_kappa







import os
import sys
from pathlib import Path

from . import read
from . import token_corpus

import spacy
from spacy.tokenizer import Tokenizer
from spacy.lang.char_classes import ALPHA, ALPHA_LOWER, ALPHA_UPPER, CONCAT_QUOTES, LIST_ELLIPSES, LIST_ICONS, HYPHENS
from spacy.util import compile_infix_regex

import re
import copy
import numpy as np


# regular expression of an integer number
# NUMBER_RE = re.compile("^[0-9]*[1-9][0-9]*$")
NUMBER_RE = re.compile("^[0-9]+$")

LINK_TYPES = ["SEMROLE", "SRLINK"]

LINK_TYPES = ["SEMROLE","SRLINK"]

infix_re = re.compile(r'[-]')


def custom_tokenizer(nlp):
    infixes = (
        LIST_ELLIPSES
        + LIST_ICONS
        + [
            r"(?<=[0-9])[+\-\*^](?=[0-9-])",
            r"(?<=[{al}{q}])\.(?=[{au}{q}])".format(
                al=ALPHA_LOWER, au=ALPHA_UPPER, q=CONCAT_QUOTES
            ),
            r"(?<=[{a}]),(?=[{a}])".format(a=ALPHA),
            r"(?<=[{a}])(?:{h})(?=[{a}])".format(a=ALPHA, h=HYPHENS),
            r"(?<=[{a}0-9])[:<>=/](?=[{a}])".format(a=ALPHA),
        ]
    )

    infix_re = compile_infix_regex(infixes)

    return Tokenizer(nlp.vocab, prefix_search=nlp.tokenizer.prefix_search,
                                suffix_search=nlp.tokenizer.suffix_search,
                                infix_finditer=infix_re.finditer,
                                token_match=nlp.tokenizer.token_match,
                                rules=nlp.Defaults.tokenizer_exceptions)

class ReadBrat(read.Read):
    """
      Reader to brat file annotations and their text files
    """

    def __init__(self, lang="en"):
        """
        Load spacy model to read process brat files. Also
        store the files processed.
        """
        if lang == "pt":
            if not(spacy.util.is_package('pt_core_news_lg')):
                spacy.cli.download('pt_core_news_lg')


            self.nlp = spacy.load("pt_core_news_lg")


            self.nlp.tokenizer = custom_tokenizer(self.nlp)
        else:
            if not (spacy.util.is_package('en_core_web_lg')):
                spacy.cli.download('en_core_web_lg')

            self.nlp = spacy.load('en_core_web_lg')

        self.file_lst = []
        self.ann_ref = {}

    def toColumn(self, data_dir, output_dir):
        """
        Convert a set of files to column format, similar to Conll

        @param string: path of data to gather and process
        @param string: path of the ouput directory

        @return None
        """

        for dirpath, dirnames, filenames in os.walk(data_dir):
            for f in filenames:
                if f.endswith(".ann"):
                    p = Path(f)
                    fullname = os.path.join(data_dir, p.stem)
                    output_file = os.path.join(output_dir, "%s.conll" % p.stem)
                    self.fileToColumnFormat(fullname, output_file)

    def _is_adjacent_offset(self, i1, i2):
        """
        test if two intervals are adjacent

        @param (int,int): a tuple of integers
        @param (int,int): a tuple of integers

        @return int: 1 if i1 is adjacent, but before i2, -1 if i1 is adjacent
        , but after i2, and 0 otherwise
        """

        s1, e1 = i1
        s2, e2 = i2

        # if the interval i1 is adjacent, but it is placed before
        # the interval i2
        if e1 + 1 == s2:
            return 1
        elif e2 + 1 == s1:
            # if the interval i1 is adjacent, but it is placed after
            # the interval i2
            return -1
        else:
            return 0

    def _get_left_span(self, ann):
        """
        Given an annotation element (dictionary), it takes the left most span interval,

        """

        # (start, end)
        return (ann["offset1"][0], ann["offset1"][1])

    def _get_right_span(self, ann):
        """
        Given an annotation element (dictionary), it takes the right most span interval,

        """
        if "offset2" in ann:
            return (ann["offset2"][0], ann["offset2"][1])
        else:
            return (ann["offset1"][0], ann["offset1"][1])

    def _is_adjacent(self, el1, el2):
        """
        Check if two annotation elements (dictionaries) are adjacent to each other
        1, if el1 < el2, -1, if el1 > el2, and 0 if el1 is not adjacent to el2
        """

        if "offset2" in el1 and "offset2" in el2:
            return 0

        (left1_start, left1_end) = self._get_left_span(el1)
        (left2_start, left2_end) = self._get_left_span(el2)


        (right1_start, right1_end) = self._get_right_span(el1)
        (right2_start, right2_end) = self._get_right_span(el2)

        ans = self._is_adjacent_offset((left1_start, left1_end), (left2_start, left2_end))
        if ans != 0: return ans

        ans = self._is_adjacent_offset((left1_start, left1_end), (right2_start, right2_end))
        if ans != 0: return ans

        ans = self._is_adjacent_offset((right1_start, right1_end), (left2_start, left2_end))
        if ans != 0: return ans

        ans = self._is_adjacent_offset((right1_start, right1_end), (right2_start, right2_end))
        if ans != 0: return ans

        return 0

    def _build_merged_element(self, el1, el2):

        new_el = {}
        if "offset2" in el1:

            new_el["offset1"] = (el1["offset1"][0], el1["offset1"][1])

            # the union if make in the second offset
            new_el["offset2"] = (el1["offset2"][0], el2["offset2"][1])

        else:
            # the union if make in the first offset
            new_el["offset1"] = (el1["offset1"][0], el2["offset1"][1])
            if "offset2" in el2:
                new_el["offset2"] = (el2["offset2"][0], el2["offset2"][1])

        new_el["value"] = el1["value"] + " " + el2["value"]
        new_el["id"] = el1["id"]
        return new_el

    def merge_span(self, ann_entity):
        """
        It merge spans of annotations

        @param [dict]: a list of annotations as dictionaries

        @param [dict]:a list of annotations as dictionaries that were merged
        if they are in the same span text
        """

        # TODO: a recursive solution will be more suitable solution

        new_ann_type = []
        merged_indexes = []  # register elements that were already merged

        for idx_el, el in enumerate(ann_entity):

            # search for an element that is inside the
            # current span or contains the current span
            ans = 0
            for idx in range(idx_el + 1, len(ann_entity)):

                ans = self._is_adjacent(el, ann_entity[idx])

                if ans > 0:
                    new_el = self._build_merged_element(el, ann_entity[idx])
                    new_ann_type.append(new_el)
                    merged_indexes.append(idx)
                    break
                elif ans < 0:
                    new_el = self._build_merged_element(ann_entity[idx], el)
                    new_ann_type.append(new_el)
                    merged_indexes.append(idx)
                    break

            if ans == 0:
                if idx_el not in merged_indexes:
                    new_ann_type.append(el)

        return new_ann_type

    def merge_entity_span(self, ann):
        """
        Given a dictionary returned by read_annotation_file method,
        it merge span of entities

        @param dict: a dictionary of different entities that contains their
        annotation

        @return dict: a dictionary of different entities that contains their
        annotation merged if they are in the same text span
        """

        new_ann = {}
        for ent_type in ann:
            if ent_type in LINK_TYPES:
                new_ann[ent_type] = ann[ent_type]
                continue # if it is a link, just ignore it
            #if "offset2_end" not in ann[ent_type]:
            #    print("-->", ann[ent_type])
            new_ann_ent = self.merge_span(ann[ent_type])
            new_ann[ent_type] = new_ann_ent

        return new_ann

    def get_attribute_value(self, element_id, attr_name):
        if attr_name in self.ann_ref:
            if element_id in self.ann_ref[attr_name]:
                return self.ann_ref[attr_name][element_id]

    def get_relation(self, element_a, element_b):
        """
        Get the relation name between two entities (their ids)
        @param element_a: String with the id of an entity
        @param element_b: String with the id of an entity
        @return: String with relation name, NOne if there is no relation between
        elements A and B
        """

        annotation_types = list(self.ann_ref.keys())
        annotation_links_types = [ann_type for ann_type in annotation_types if "LINK" in ann_type]

        for relation_name in  annotation_links_types:
            if element_a in self.ann_ref[relation_name]:
                if element_b in self.ann_ref[relation_name][element_a]:
                    return relation_name



    def read_annotation_file(self, file_ann, merge_entities=False):
        """
        It reads only the annotation file, then returns
        the processed tokens as TokenCorpus type

        @param string: path of data to gather and process

        @return dictionary: a dictionary of annotations
        """

        ann = {"Event": [], "Actor": [], "Time": [], "TIME_X3": [], \
                "ACTOR": [], "Participant": [],"SRLINK":[],"OLINK":[],"TLINK":[]}
        ann_ref = {}  # annotation by reference
        # the pattern of a reference for an argument of a relation
        ARG_REL = re.compile("T\d+")

        with open(file_ann, "r") as fd:
            last_line_type = None
            for line in fd:
                ann_type = None
                line_toks = line.split()
                if line[0] != '#':

                    if len(line_toks) < 3:
                        if line[0] != 'T' and line[0] != 'A' and line[0] != 'R' and last_line_type in ann.keys():
                            ann[last_line_type][-1]["value"] += " " + line.strip()
                        else:
                            raise Exception(f"Invalid standoff line:{line}")
                        continue
                    ann_type = line_toks[1]
                    last_line_type = ann_type

                    if line[0] == 'T':
                        # TODO: catalogar aqui todos as entidades em um mapeamento

                        if ann_type not in ann:
                            ann[ann_type] = []

                        if ';' in line_toks[3]:

                            start_range = int(line_toks[2])
                            offset_lst = []

                            idx_toks = 3
                            while ';' in line_toks[idx_toks]:
                                tmp = line_toks[idx_toks].split(';')
                                end_range = int(tmp[0])
                                offset_lst.append((start_range, end_range))

                                start_range = int(tmp[1])
                                idx_toks += 1

                            end_range = int(line_toks[idx_toks])
                            offset_lst.append((start_range, end_range))

                            segment_ann = {"id":line_toks[0]} # annotation multi-segment annotation
                            for idx_offset, (start_range, end_range) in enumerate(offset_lst):
                                segment_ann[f"offset{idx_offset}"] = (start_range, end_range)

                            value = " ".join(line_toks[idx_toks + 1:])
                            segment_ann["value"] = value
                            ann[ann_type].append(segment_ann)

                        else:
                            offset1_start = int(line_toks[2])
                            offset1_end = int(line_toks[3])

                            value = " ".join(line_toks[4:])

                            ann[ann_type].append({"id":line_toks[0],"offset1": (offset1_start, offset1_end), \
                                              "value": value})
                    elif line[0] == 'R':
                        e1 = line_toks[2].split(":")[1]  # entity 1
                        e2 = line_toks[3].split(":")[1]  # entity 2
                        rel_id = line_toks[0]
                        if ann_type in ann:
                            ann[ann_type].append({"id":rel_id,"args":(e1, e2)})
                        else:
                            ann[ann_type] = [{"id": rel_id, "args": (e1, e2)}]


                        if ann_type in self.ann_ref:
                            if e1 in self.ann_ref[ann_type]:
                                self.ann_ref[ann_type][e1].append(e2)
                            else:
                                self.ann_ref[ann_type][e1] = [e2]
                        else:
                            self.ann_ref[ann_type] = {e1: [e2]}

                    elif line[0] == 'A':

                        ref = line_toks[2]
                        if ref in ann_ref:
                            # ann_ref[ref][1].append((line[1], line[3]))
                            if len(line_toks) > 3:
                                ann_ref[ref][line_toks[1]] = line_toks[3]
                            else:
                                ann_ref[ref][line_toks[1]] = line_toks[1]

                        else:
                            ann_ref[ref] = {line_toks[1]: line_toks[3]}
                    elif line[0] == '#':
                            # it is a note about the annotation,
                            # insert it as an attribute
                            ann_type = line_toks[1]
                            ref = line_toks[2]
                            ann_info = " ".join(line_toks[3:])

                            if ref in ann_ref:
                                if ann_type in ann_ref[ref]:
                                    ann_ref[ref][ann_type].append(ann_info)
                                else:
                                    ann_ref[ref][ann_type] = [ann_info]
                            else:
                                ann_ref[ref] = {ann_type: [ann_info]}

        if merge_entities:
            return self.merge_entity_span(ann)
        else:
            return ann

    def process(self, data_dir, split=None):
        """
        It reads a set of files of annotations and text, then returns
        the processed tokens as TokenCorpus type

        @param string: path of data to gather and process
        @param string: a file name that contains a list of the files in data dir that
        should be processed by this reader

        @return [[TokenCorpus]]: a list of lists of tokens
        """
        # process the data corpus
        # and return a list of tokens
        split_lst = []
        if split is not None:
            with open(split, "r") as fd:
                split_lst = fd.readlines()
                split_lst = [line.replace("\n","") for line in split_lst]

        data_tokens = []

        for dirpath, dirnames, filenames in os.walk(data_dir):
            for f in filenames:

                if f.endswith(".ann"):

                    p = Path(f)
                    if split_lst != [] and not(p.stem + ".txt" in split_lst):
                        continue
                 
                    fullname = os.path.join(data_dir, p.stem)
                    #print("-->", fullname)
                    token_lst = self.process_file(fullname)
                    self.file_lst.append(fullname + ".txt")

                    if len(token_lst) > 0:
                        data_tokens.append(token_lst)

        return data_tokens

    def extract_token_corpus(self, doc, ann_idx, ann, ann_ref):
        """
        Build a list of tokens using TokenCorpus object. It uses 
        a document of spacy.Document type (doc) and a list of indexes.

        ann_idx is a maps a 2-tuple integer to a list of annotation. Regarfing, 
        the 2-tuple, the first element  is the start character offset of 
        an annotated token in the 
        raw text of the document, and the second element is end of the character 
        offset of an annotated token.

        ann is a dictionary that maps a 2-tuple of indexes to a list of 
        the annotation of the given span indexes. 

        ann_ref is a dictionary that maps a reference (i.e. an id of an annotation 
        assigned by BRAT) to the attribute list of an annotation span
        """
        count = 0
        token_lst = []
        ref2tok = {}

    
        sent_id = -1
        clause_id = 0
        for tok_idx, tok in enumerate(doc):
            if tok.is_sent_start:
                sent_id += 1

            # Check if the token marks the end of a clause
            if tok.dep_ in ["punct", "mark"]:
                clause_id += 1

            
            mytok = token_corpus.TokenCorpus()
            mytok.id = count
            mytok.text = tok.text
            mytok.lemma = tok.lemma_
            mytok.pos = tok.pos_
            mytok.dep = tok.dep_
            mytok.head = tok.head.text
            mytok.head_pos = tok.head.pos_
            mytok.head_lemma = tok.head.lemma_
            mytok.offset = tok.idx
            mytok.sent_id = sent_id
            mytok.clause_id = clause_id
            #print("-->",clause_id, sent_id)


            ans = self.search_all_idx(tok.idx, tok.idx + len(tok.text), ann_idx)

            # TODO: it is necessary perform a more efficient search
            # a possible subtoken annotation
            # perform more than one search, and build a list of id's
            if len(tok.text.strip()) != 0 and tok.text not in ".!?-,":
                ans_sub = self.search_subtoken(tok.idx, tok.idx + len(tok.text) - 1, \
                        ann_idx)

                ans = ans.union(ans_sub)

            if len(ans) != 0:
                # annotations in token
                # a token can be annotated twice, and besides that,
                # can be a part of a span annotation..The iteration in ans
                # tries to get the span annotations in one token, and the 
                # iteration in ann[(t0,t1)] bellow tries to get several annotations
                # in only one token
                ref_lst_tok = set()
                for idx in ans:
                    (t0, t1) = ann_idx[idx]


                    for ref, ann_type, elems in ann[(t0, t1)]:

                        mytok.attr.append((ann_type,ann_ref[ref]))
                        mytok.ann_offset.append((t0,t1))
                        mytok.id_ann.append(ref) 
                        # all the annotations ids that a token can be assigned
                        ref_lst_tok.add(ref)


                # now update the mapping of references to tokens with
                # the token object that was build in this iteration
                for ref in ref_lst_tok:
                    if ref in ref2tok:
                        ref2tok[ref].append(mytok)
                    else:
                        ref2tok[ref] = [mytok]

            token_lst.append(mytok)

            count += 1

        return token_lst, ref2tok

    # TODO: refactor this method. It is a very long and confusing method
    def process_file(self, data_file):
        """
        It reads only one file of annotation and text, then returns
        the processed tokens as TokenCorpus type

        @param string: path of data to gather and process

        @return [TokenCorpus]: a list of tokens
        """

        file_ann = "%s.ann" % data_file
        file_txt = "%s.txt" % data_file


        if not os.path.exists(file_txt) or not os.path.exists(file_ann):
            return []

        ann = {}  # the index is the offset of the annotation
        ann_ref = {}  # annotation by reference
        ann_rel = {} # annotation by relation

        # the pattern of a reference for an argument of a relation
        ARG_REL = re.compile("T\d+") 

        with open(file_ann, "r") as fd:

            for line in fd:
                line_toks = line.split()
                if line[0] != '#' and line[0] == 'T':

                    # ann[offset start] = (ref, ann_type, value_str)
                    if ';' in line_toks[3]:
                        # this situation is when the annotation of th event
                        # is two segments not adjacents
                        start_range = int(line_toks[2])
                        offset_lst = []

                        idx_toks = 3
                        while ';' in line_toks[idx_toks]:

                            tmp = line_toks[idx_toks].split(';')
                            end_range = int(tmp[0])
                            offset_lst.append((start_range, end_range))

                            start_range = int(tmp[1])
                            idx_toks += 1

                        end_range = int(line_toks[idx_toks])
                        offset_lst.append((start_range, end_range))

                        for start_range, end_range in offset_lst:
                            if (start_range, end_range) in ann:
                                ann[(start_range, end_range)].append((line_toks[0], line_toks[1], line_toks[idx_toks + 1:]))
                            else:
                                ann[(start_range, end_range)] = [(line_toks[0], line_toks[1], line_toks[idx_toks + 1:])]

                        #ann[(int(line_toks[2]), int(tmp[0]))] = [(line_toks[0], line_toks[1], line_toks[5:])]
                        #ann[(int(tmp[1]), int(line_toks[4]))] = [(line_toks[0], line_toks[1], line_toks[5:])]
                        #ann_ref[line[0]] = (
                        #    [(int(line_toks[2]), int(tmp[0])), (int(tmp[1]), int(line_toks[4]))], [])
                        ann_ref[line[0]] = (offset_lst, [])


                        ann_ref[line_toks[0]] = {}
                    else:
                        if NUMBER_RE.match(line_toks[2]):

                            offset_start, offset_end = int(line_toks[2]), int(line_toks[3])

                            if (offset_start, offset_end) in ann:
                                ann[(offset_start, offset_end)].append((line_toks[0], line_toks[1], line_toks[4:]))
                            else:
                                ann[(offset_start, offset_end)] = [(line_toks[0], line_toks[1], line_toks[4:])]
                                
                            ann_ref[line_toks[0]] = {}
                elif line[0] == 'R':
                    rel_id = line_toks[0]

                    rel_type = line_toks[1]
                    ref_lst = ARG_REL.findall(line)

                    if len(ref_lst) > 1:

                        ref1 = ref_lst[0]
                        ref2 = ref_lst[1]

                        if ref1 in ann_rel:
                            ann_rel[ref1].append((rel_id, rel_type, ref2, "arg2"))
                        else:
                            ann_rel[ref1] = [(rel_id, rel_type, ref2,"arg2")]

                        if ref2 in ann_rel:
                            ann_rel[ref2].append((rel_id, rel_type, ref1,"arg1"))
                        else:
                            ann_rel[ref2] = [(rel_id, rel_type, ref1,"arg1")]

                    else:
                        print("Warning: There is a relation with only one argument.")



                elif line[0] == 'A':

                    ref = line_toks[2]
                    if ref in ann_ref:
                        #ann_ref[ref][1].append((line[1], line[3]))
                        if len(line_toks) > 3:
                            ann_ref[ref][line_toks[1]] = line_toks[3]
                        else:
                            ann_ref[ref][line_toks[1]] = line_toks[1]

                    else:
                        ann_ref[ref] = {line_toks[1]:line_toks[3]}
                elif line[0] == '#':
                    # it is a note about the annotation,
                    # insert it as an attribute
                    ann_type = line_toks[1]
                    ref = line_toks[2]
                    ann_info = " ".join(line_toks[3:])

                    if ref in ann_ref:
                        if ann_type in ann_ref[ref]:
                            ann_ref[ref][ann_type].append(ann_info)
                        else:
                            ann_ref[ref][ann_type] = [ann_info]
                    else:
                        ann_ref[ref] = {ann_type: [ann_info]}

        idx_lst = ann.keys()

        idx_lst = sorted(idx_lst, key=lambda elem: elem[0])

        # a mapping between ref and Tok object
        # the use it to point to the tokens that it own relations
        ref2tok = {} 
        token_lst = []

        with open(file_txt, "r") as fd:

            doc = self.nlp(fd.read().strip())
            token_lst, ref2tok = self.extract_token_corpus(doc, idx_lst, ann, ann_ref)

        # update relation field for each token
        # each relation field is a list of TokenRelation object, which 
        # points to a TokenCorpus and specifies the type of relation
        for mytok in token_lst:

            if mytok.id_ann != []:

                for ref  in mytok.id_ann:

                    if ref in ann_rel: # if the current token has any relation, then...

                        # the relation for the reference ref
                        for rel_id, rel_type, ref_arg,argn in ann_rel[ref]:
                            try:
                                rel_obj = token_corpus.TokenRelation(rel_id, ref2tok[ref_arg], rel_type, argn, ref_arg)
                                mytok.relations.append(rel_obj)
                            except KeyError:
                                print(f"Warning: The {rel_id} was not included in the doc since {ref_arg} was not found as annotation.")

        return token_lst

    def search_all_idx(self, idx, idx_end, idx_lst):
        """
        Since a token can be annotated with multiple id's, it is necessary 
        to perform multiple search for each one of these id's.
        """

        all_idx_lst = set()

        copy_idx_lst = [e for e in idx_lst]
        
        while True:

            ans = self.search_idx(idx, copy_idx_lst)
            if ans == -1:
                return all_idx_lst
            else:
                # update index according to the position
                # in the idx_lst 
                elem = copy_idx_lst[ans]
                copy_idx_lst.pop(ans)
                if len(idx_lst) > 0:
                    # can I afford such expensive operation
                    # since few of them are possible performed?
                    ans = idx_lst.index(elem)

                all_idx_lst.add(ans)


    def search_subtoken(self, tok_start, tok_end, idx_lst):

        ans = set()

        for pos, (start, end) in enumerate(idx_lst):
            
            #if start >= tok_start and end <= tok_end:
            if not (tok_end < start or tok_start > end):
                ans.add(pos)

        return ans

    def search_idx(self, idx, idx_lst):
        """
        it searches for tuples (t0, t1) in idx_lst where idx >= t0 and idx <= t1

        @param integer: an index
        @param [(integer,integer)]: a list of index

        @return integer: the position of the tuple or -1 if none is found
        """

        b = 0
        e = len(idx_lst) - 1

        m = int((b + e) / 2)
        pos = -1

        while b <= e:
            (t0, t1) = idx_lst[m]

            if idx == t0:
                pos = m
                break
            elif idx > t0:
                if idx >= t1:
                    b = m + 1
                else:
                    pos = m
                    break
            else:
                e = m - 1

            m = int((b + e) / 2)

        return pos

    def __process_annotations(self, file_ann, ann, ann_ref):
        pass


    def fileToColumnFormat(self, ann_file, output_file):
        """
        Convert only one file to the column format.

        @param string: path of annotation file
        @param string: path of output file

        @return None
        """
        print("Processing %s..." % ann_file)

        tok_lst = self.process_file(ann_file)
        with open(output_file, "w") as fd:

            for tok in tok_lst:

                if tok.ann == 'Event':
                    ann = 'I'
                else:
                    ann = 'O'

                fd.write("%d %s %s %s\n" % (tok.id, tok.text.strip(), tok.pos, ann))
        print("Output %s" % output_file)


if __name__ == "__main__":
    # only to unit  tests
    data_dir = os.environ.get("DATA_DIR")
    output_dir = os.environ.get("COLUMN_DIR")

    if data_dir is None:
        print("Please, set DATA_DIR enviroment variable.")
        sys.exit(0)
    if output_dir is None:
        print("Please, set COLUMN_DIR enviroment variable.")
        sys.exit(0)

    r = ReadBrat()
    # r.process(data_dir) # read and return a list of list of tokens
    r.toColumn(data_dir, output_dir)



import json
import sys
import os

from evaluation import start, build_evaluation, write_metrics_result



def read_config(config_file_name):

    with open(config_file_name, "r") as fd:
        config_exp = json.load(fd)
    return config_exp

def run():

    config_file_name = sys.argv[1]

    config_exp = read_config(config_file_name)

    for name_experiment in config_exp:

        if not(os.path.exists(name_experiment)):
            os.mkdir(name_experiment)

        language = config_exp[name_experiment]["language"]
        start(language)
        tools = config_exp[name_experiment]["tools"]

        narrative_elements = {}
        if "participant" in tools:
            narrative_elements["participant"] = tools["participant"]
        if "time" in tools:
            narrative_elements["time"] = tools["time"]
        if "event" in tools:
            narrative_elements["event"] = tools["event"]
        if "srlink" in tools:
            narrative_elements["srlink"] = tools["srlink"]

        data_dir = config_exp[name_experiment]["data"]
        split_info = config_exp[name_experiment].get("split", None)

        res = build_evaluation(narrative_elements=narrative_elements, language=language, data_dir=data_dir,
                           results_dir=name_experiment,split=split_info)

        with open("%s_metrics.txt" % name_experiment, "w") as fd:
            write_metrics_result(res, fd)


if __name__ == "__main__":
    run()



from bs4 import BeautifulSoup
import sys
import requests
import socket

from text2story.core.exceptions import InvalidLanguage, NoConnection


# Language mapping for DBpedia Spotlight API URL
language_to_url = {
    'en': 'http://api.dbpedia-spotlight.org/en/annotate',
    'pt': 'http://api.dbpedia-spotlight.org/pt/annotate'
}

def load(lang, key=None):
    """
    Used, at start, to load the pipeline for the supported languages.
    """
    # Checks if the language is supported
    if lang not in language_to_url:
        raise InvalidLanguage(lang)

def check_internet_connection(host="8.8.8.8", port=53, timeout=3):
    """
    Check internet connection by trying to connect to a specified host and port.
    """
    try:
        # Create a socket object
        socket.setdefaulttimeout(timeout)
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        # Try to connect to the host and port
        s.connect((host, port))
        s.close()
        return True
    except OSError:
        return False


def extract_participants(text, lang=None, url=None):

    # Request parameters
    params = {
        "text": text,
        "confidence": 0.2,  # Minimum confidence level for annotations
        "support": 20       # Minimum number of supporting documents
    }

    if lang:
        # Check internet connection
        if not check_internet_connection():
            raise NoConnection

        # Checks if the language is supported
        if lang not in language_to_url:
            raise InvalidLanguage(lang)

        # DBpedia Spotlight API URL based on language
        url = language_to_url[lang]

    elif url:
        # DBpedia Spotlight service URL in Docker container
        url = url

        # Request parameters
        params = {
            "text": text,
            "confidence": 0.2,  # Minimum confidence level for annotations
            "support": 20       # Minimum number of supporting documents
        }

        # Making the POST request
        response = requests.post(url, data=params)
    else:
        raise BaseException(f"Error calling extract_participants. Check if you wrote all the necessary parameters in the function.")
        
    # Making the GET request
    response = requests.get(url, params=params)

    # Checks whether the request was successful
    if response.status_code == 200:
        # Parse the HTML returned by the response
        soup = BeautifulSoup(response.content, 'html.parser')
        
        # Finds all <a> tags that contain information about resources
        resource_tags = soup.find_all('a')

        # Extract information from resources
        participants = []
        for tag in resource_tags:
            actor = tag.get_text()
            start_character_offset = params['text'].find(str(actor))
            end_character_offset = start_character_offset + len(str(actor))
            actor_head = None
            actor_type = 'Other'
            participants.append(((start_character_offset, end_character_offset), actor_head, actor_type))

        return participants
    else:
        raise BaseException(f"Error calling DBpedia URL {url}. Status code: {response.status_code}")




import os
import sys

from pathlib import Path

import text2story as t2s  # Import the package
from text2story.readers import read_brat

import argparse
import functools
from datetime import datetime

from text2story.experiments.metrics import *


def start(lang):
    t2s.start(lang)  # Load the pipelines


##########################
### Auxiliary methods ####
##########################

def dir_path(path):
    if os.path.isdir(path):
        return path
    else:
        raise argparse.ArgumentTypeError(f"readable_dir:{path} is not a valid path")


#########################################
### Prediction and evaluation methods ###
#########################################

def extract_element(doc, el, tool):
    """
    It extracts a given element (participant, event or time) using a given
    tool.

    @param doc: a Narrative object
    @param string: a string specifying the type of element to be extracted
    @param string: a string specifying the type of tool to apply in the document

    @return None
    """

    if el == 'participant':
        doc.extract_participants(tool)
    elif el == 'event':
        doc.extract_events(tool)
    elif el == 'time':
        doc.extract_times(tool)
    elif el == 'srlink':
        doc.extract_semantic_role_links(tool)
    elif el == 'olink':
        doc.extract_objectal_links(tool)
    else:
        raise Exception("extract_element: Unrecognize element %s" % el)


def evaluate_element(pred_file, target_file, el):
    """
    It evaluates a given element (participant, event or time) .

    @param string: path of prediction file (.ann)
    @param string: path of target annotated file (.ann)
    @param string: a string specifying the type of elment to evaluate

    @return (dict, dict) the relaxed and strict results
    """

    if el == 'participant':
        return evaluate_actor(pred_file, target_file)
    elif el == 'event':
        return evaluate_event(pred_file, target_file)
    elif el == 'time':
        return evaluate_time(pred_file, target_file)
    elif el == 'srlink':
        return evaluate_srlink(pred_file, target_file)
    else:
        raise Exception("evaluate_element: Unrecognize element %s" % el)


def get_element(elem, ann_pred, ann_target):
    if elem == "participant":
        return get_element_actor(ann_pred, ann_target)
    elif elem == "event":
        return get_element_event(ann_pred, ann_target)
    elif elem == "time":
        return get_element_time(ann_pred, ann_target)
    elif elem == "srlink":
        return get_element_srlink(ann_pred, ann_target)
    else:
        raise Exception("get_element: Unrecognize element %s" % elem)


def get_element_srlink(ann_pred, ann_target):
    # aqui deveria retorna um dicionario  e nao uma lista, visto que existem varios tipos
    # de links. Por outro lado, para cada tipo, devem existir em pouca quantidade. 
    # o ideal é quantificar cada tipo de srlink. Dai analisar se vale a pena fazer a análise por
    # tipo aqui
    # r += (f"{sem_link_id}\tSEMROLE_{sem_link.type} Arg1:{sem_link.event} Arg2:{sem_link.actor}\n")
    # r += (f"{sem_link_id}\tSRLINK_{sem_link.type} Arg1:{sem_link.event} Arg2:{sem_link.actor}\n")
    srlink_pred = []
    srlink_target = []

    for k in ann_pred:
        if k.startswith("SEMROLE") or k.startswith("SRLINK"):
            srlink_pred = ann_pred[k]

    for k in ann_target:
        if k.startswith("SEMROLE") or k.startswith("SRLINK"):
            srlink_target = ann_target[k]

    actors_pred, actors_target = get_element_actor(ann_pred, ann_target)
    event_pred, event_target = get_element_event(ann_pred, ann_target)

    pred = (actors_pred, event_pred, srlink_pred)
    target = (actors_target, event_target, srlink_target)
    return pred, target


def get_element_event(ann_pred, ann_target):
    # compute accuracy of the exacttly same span
    event_pred = ann_pred["Event"]
    event_target = ann_target["Event"]

    return event_pred, event_target


def get_element_actor(ann_pred, ann_target):
    actor_pred = ann_pred["Actor"]
    # conditions to preserve compatibility between anotation versions
    if len(actor_pred) == 0:
        actor_pred = ann_pred["ACTOR"]
    if len(actor_pred) == 0:
        actor_pred = ann_pred["Participant"]

    actor_target = ann_target["Actor"]
    if len(actor_target) == 0:
        actor_target = ann_target["ACTOR"]
    if len(actor_target) == 0:
        actor_target = ann_target["Participant"]

    return actor_pred, actor_target


def get_element_time(ann_pred, ann_target):
    time_pred = ann_pred["Time"]
    if len(time_pred) == 0:
        time_pred = ann_pred["TIME_X3"]
    time_target = ann_target["Time"]

    return time_pred, time_target


def evaluate_event(pred_file, target_file):
    """
    Implements token event precision/recall/f1 and span event precision/recall/f1

    @param string: predict file in the brat .ann format
    @param string: target file (human labeled) in the brat .ann format

    @return a tuple of dictionaries
    """

    reader = read_brat.ReadBrat()

    ann_pred = reader.read_annotation_file(pred_file)
    ann_target = reader.read_annotation_file(target_file)

    # compute accuracy of the exacttly same span
    event_pred = ann_pred["Event"]
    event_target = ann_target["Event"]

    scores_relax = compute_relax_scores("event", event_pred, event_target)
    scores = compute_strict_scores("event", event_pred, event_target)
    return scores_relax, scores


def evaluate_actor(pred_file, target_file):
    """
    Implements token actor precision/recall/f1 and span actor precision/recall/f1

    @param string: predict file in the brat .ann format
    @param string: target file (human labeled) in the brat .ann format

    @return a tuple of dictionaries
    """

    reader = read_brat.ReadBrat()

    ann_pred = reader.read_annotation_file(pred_file)
    ann_target = reader.read_annotation_file(target_file)

    # compute accuracy of the exacttly same span

    actor_pred = ann_pred["Actor"]
    # conditions to preserve compatibility between anotation versions
    if len(actor_pred) == 0:
        actor_pred = ann_pred["ACTOR"]
    if len(actor_pred) == 0:
        actor_pred = ann_pred["Participant"]

    actor_target = ann_target["Actor"]
    if len(actor_target) == 0:
        actor_target = ann_target["ACTOR"]
    if len(actor_target) == 0:
        actor_target = ann_target["Participant"]

    scores_relax = compute_relax_scores("participant", actor_pred, actor_target)
    scores = compute_strict_scores("participant", actor_pred, actor_target)
    return scores_relax, scores


def evaluate_time(pred_file, target_file):
    """
    Implements token time precision/recall/f1 and span time precision/recall/f1

    @param string: predict file in the brat .ann format
    @param string: target file (human labeled) in the brat .ann format

    @return a tuple of dictionaries
    """

    reader = read_brat.ReadBrat()
    ann_pred = reader.read_annotation_file(pred_file)
    ann_target = reader.read_annotation_file(target_file)

    # compute accuracy of the exacttly same span
    time_pred = ann_pred["Time"]
    if len(time_pred) == 0:
        time_pred = ann_pred["TIME_X3"]
    time_target = ann_target["Time"]

    scores_relax = compute_relax_scores("time", time_pred, time_target)
    scores = compute_strict_scores("time", time_pred, time_target)
    return scores_relax, scores


def evaluate_srlink(pred_file, target_file):
    """
    Implements token srlink precision/recall/f1 and span time precision/recall/f1

    @param string: predict file in the brat .ann format
    @param string: target file (human labeled) in the brat .ann format

    @return a tuple of dictionaries
    """

    reader = read_brat.ReadBrat()
    ann_pred = reader.read_annotation_file(pred_file)
    ann_target = reader.read_annotation_file(target_file)

    # compute accuracy of the exacttly same span
    pred, target = get_element("srlink", ann_pred, ann_target)

    scores_relax = compute_relax_scores("srlink", pred, target)
    scores = compute_strict_scores("srlink", pred, target)
    return scores_relax, scores


def prediction(input_dir, results_dir, narrative_elements, language, split=None):
    """
    Read brat data (.ann and .txt files) in the input directory,
    and write the results (columns files: token, pred_label, target_label)
    in the results dir

    @param string: input directory with ann and txt files
    @param string: directory with results file for each document file in
    the input_dir

    @return [(string,string)]: a tuple file list to compare
    """

    reader = read_brat.ReadBrat()
    doc_lst = reader.process(input_dir, split)

    doc_pred_target = []  #

    for idx_doc, doc in enumerate(doc_lst):

        ann_filename = os.path.basename(reader.file_lst[idx_doc])
        ann_filename = os.path.join(results_dir, ann_filename)

        target_file = Path(reader.file_lst[idx_doc]).stem + ".ann"
        target_file = os.path.join(input_dir, target_file)

        if os.path.exists(ann_filename):
            doc_pred_target.append((ann_filename, target_file))
            print(f"{ann_filename} annotation already exists. Skipping it..")
            continue

        text_ = ""
        with open(reader.file_lst[idx_doc], "r") as fd:
            text_ = fd.read()

        narrative_doc = t2s.Narrative(language, text_, "2020-10-11")

        # extract the element in the given tool
        for el in narrative_elements:
            print("%s extracting from file %s" % (narrative_elements[el], reader.file_lst[idx_doc]))
            extract_element(narrative_doc, el, narrative_elements[el])

        iso_str = narrative_doc.ISO_annotation()
        with open(ann_filename, "w") as fd:
            fd.write(iso_str)

        doc_pred_target.append((ann_filename, target_file))

    return doc_pred_target


def process_evaluation(narrative_elements, doc_lst):
    """
    Process evaluation for a given element (time, actors or event), in a
    given tool (spacy, spacy, py_heideltime, custompt, etc).

    You should set DATA_DIR and RESULTS_DIR in your PATH enviroment
    vari

    @param string: the element to be extracted
    @param string: the annotator tool to be employed in the extraction

    @param dictionary: a dictionary with the results
    """

    res = {}
    metrics = ["precision_relax", "recall_relax", "f1_relax", \
               "precision", "recall", "f1"]

    for elem in narrative_elements:
        for m in metrics:
            res[m + "_" + elem] = []

    reader = read_brat.ReadBrat()

    for pred_file, target_file in doc_lst:
        print("Evaluating %s and %s" % (pred_file, target_file))


        ann_pred = reader.read_annotation_file(pred_file, merge_entities=True)
        ann_target = reader.read_annotation_file(target_file, merge_entities=False)

        for elem in narrative_elements:
            pred, target = get_element(elem, ann_pred, ann_target)

            # TODO: esse e so para o srlink. Depois posso querer
            # fazer a divisao por tipos
            # if len(pred) > 0 and len(target) > 0 and \
            #        type(pred[0]) == type(tuple()):
            #    pred = functools.reduce(lambda x,y:x+y,\
            #            [score for (type_pred, score) in pred])
            #    target = functools.reduce(lambda x,y:x+y,\
            #            [score for (type_target, score) in target])

            try:
                precision_relax, recall_relax, f1_relax = compute_relax_scores(elem, pred, target)
                precision, recall, f1 = compute_strict_scores(elem, pred, target)

                res["precision_relax_" + elem].append(precision_relax)
                res["recall_relax_" + elem].append(recall_relax)
                res["f1_relax_" + elem].append(f1_relax)

                res["precision_" + elem].append(precision)
                res["recall_" + elem].append(recall)
                res["f1_" + elem].append(f1)
            except:
                e = sys.exc_info()[0]
                with open("log_evaluation", "a") as fd_log:
                    now = datetime.now()
                    time_str = now.strftime("%m/%d/%Y, %H:%M:%S")
                    fd_log.write("[%s] <p>Error: %s</p>\n" % (time_str, e))
                print("Warning: Some error computing score of %s file" % pred_file)

    return res


def build_evaluation(narrative_elements, language, data_dir: str, results_dir: str, split=None):
    """
    Process the evaluation of DATA_DIR (enviroment variable) and put
    the extracted elements in the RESULTS_DIR (enviroment variable)

    @param dict: A dictionary with the elements (actor, time, event) with the
    the tool list to be employed
    @param string: the language to be evaluated (pt or en)
    @param: The System Path where data is stored
    @parm: The System Path where results will be output
    @return None
    """

    doc_pred_lst = prediction(data_dir, results_dir, narrative_elements, language, split)
    return process_evaluation(narrative_elements, doc_pred_lst)


def print_metrics_result(res: dict):
    print(f"\n-------Metrics Results-------")

    for key_tool in res.keys():
        avg_value = calculate_average_result(res[key_tool])
        print(f"Average Value for Metric {key_tool} is: {avg_value}")


def write_metrics_result(res, fd):
    fd.write(f"\n-------Metrics Results-------\n")

    for key_tool in res.keys():
        avg_value = calculate_average_result(res[key_tool])
        fd.write(f"Average Value for Metric {key_tool} is: {avg_value}\n")


def calculate_average_result(result) -> float:
    counter = 0

    for elem in result:
        counter += elem / len(result)

    return counter


def main(narrative_elements: dict, language: str, data_dir: str = None, results_dir: str = None):
    start(language)

    res = build_evaluation(narrative_elements=narrative_elements, language=language, data_dir=data_dir,
                           results_dir=results_dir)

    print_metrics_result(res)


if __name__ == "__main__":

    my_parser = argparse.ArgumentParser(description='Evaluation of a give dataset according to standard metrics')

    my_parser.add_argument("inputdir", action='store', type=dir_path,
                           help="The directory that contains the target files (brat format) and the txt narrative files.")
    my_parser.add_argument("resultsdir", action='store', type=dir_path,
                           help="The directory where are the files with the extracted entities.")

    my_parser.add_argument("--language", action='store', type=str,
                           help="Current support en (English) and pt (Portuguese. Default: en.")

    my_parser.add_argument("--participant", action='store', type=str,
                           help="The tools to extract participants from narratives. Default: spacy.")
    my_parser.add_argument("--time", action='store', type=str,
                           help="The tools to extract time from narratives. Default: py_heideltime.")
    my_parser.add_argument("--event", action='store', type=str,
                           help="The tools to extract event from narratives. Default: allennlp.")

    my_parser.add_argument("--srlink", action='store', type=str,
                           help="The tools to extract srlinks from narratives. Default: allennlp.")

    args = my_parser.parse_args()

    language = 'en'
    if args.language is not None:
        if args.language != 'en' and args.language != 'pt':
            print("Language option not recognized: %s." % args.language)
            sys.exit()
        language = args.language

    participant_tool = 'spacy'
    if args.participant is not None:
        if args.participant not in t2s.participant_tool_names():
            print("Participant tool not recognized: %s." % args.participant)
            sys.exit()
        participant_tool = args.participant

    time_tool = 'py_heideltime'
    if args.time is not None:
        if args.time not in t2s.time_tool_names():
            print("Time tool not recognized: %s." % args.time)
            sys.exit()
        time_tool = args.time

    event_tool = 'allennlp'
    if args.event is not None:
        if args.event not in t2s.event_tool_names():
            print("Event tool not recognized: %s." % args.event)
            sys.exit()
        event_tool = args.event

    srlink_tool = 'allennlp'
    if args.srlink is not None:
        if args.srlink not in t2s.srlink_tool_names():
            print("Event tool not recognized: %s." % args.event)
            sys.exit()
        srlink_tool = args.srlink

    narrative_elements = {"participant": participant_tool, \
                          "time": time_tool, \
                          "event": event_tool, \
                          "srlink": srlink_tool}

    main(narrative_elements, language, \
         args.inputdir, args.resultsdir)








class Read:
    """
    Abstract class to read all different kind of corpus
    """

    def process(self, data_dir):
        """
        Process a set of files of a corpus

        @param string: path of data corpus

        @return: a list of tokens
        """
        pass

    def process_file(self, data_file):
        """
        Process only one file of a corpus.
        """
        pass

    def __process_annotations(self, data_file):
        """
        Process specific annotations of the given data file
        """
        pass





import os

from text2story.experiments.evaluation import process_evaluation, print_metrics_result


def run():
    narrative_elements = ["event", "participant", "time"]
    #target_dir = "/home/evelinamorim/UPorto/gpt_struct_me/resources/lusa_news/"
    #pred_dir = "/home/evelinamorim/UPorto/gpt_struct_me/brat/lusa_news/"
    target_dir = "/home/evelinamorim/UPorto/text2story2019/Data/AceISO_test/"
    pred_dir = "/home/evelinamorim/UPorto/gpt_struct_me/brat/ace"

    target_files = os.listdir(target_dir)
    pred_files = os.listdir(pred_dir)

    doc_lst = []
    for p_file in pred_files:
        if p_file in target_files:
            doc_lst.append((os.path.join(pred_dir, p_file),\
                           os.path.join(target_dir, p_file)))

    res = process_evaluation(narrative_elements, doc_lst)
    print_metrics_result(res)

if __name__ == "__main__":
    run()



def get_all_participants_sent(tok_lst):
    """
    Get all participants from a document and index them by the sentence number
    @param tok_lst:
    @return:
    """

    participant_lst = {}
    participant_id_lst = set()

    for i in range(len(tok_lst)):
        tok = tok_lst[i]

        if tok.is_type("Participant") and tok.id_ann[0] not in participant_id_lst:
            if tok.sent_id in participant_lst:
                participant_lst[tok.sent_id].append(tok)
            else:
                participant_lst[tok.sent_id] = [tok]
            participant_id_lst.add(tok.id_ann[0])

    return participant_lst

def get_all_participants(tok_lst):
    participants_lst = {}

    for i in range(len(tok_lst)):
        tok = tok_lst[i]

        for attr_item in tok.attr:

            ann_type = attr_item[0]
            attr_map = attr_item[1]

            if "participant" in ann_type.lower():

                for id_ann in tok.id_ann:

                    if id_ann in participants_lst:
                        participants_lst[id_ann].append(tok)
                    else:
                        participants_lst[id_ann] = [tok]
                break

    return participants_lst


import os.path

import plantuml

from text2story.readers.read_brat import ReadBrat
from text2story.select.bubble import BubbleMap
from text2story.select.event import sieve_bbubbles_events, get_all_events
from text2story.core.utils import join_tokens

MAP_COLORS = {"State":"#D7BDE2",\
              "Reporting":"#85C1E9",
              "Occurrence":"#EDBB99",\
              "Perception":"#EC7063 ",\
              "Aspectual":"#D5DBDB",\
              "I_Action":"#48C9B0",\
              "I_State":"#F7DC6F"}

def build_fig(doc):
    msc_str = "@startuml\nskinparam style strictuml\n"

    # for each id annotation get the related events
    all_event_lst = get_all_events(doc)

    # get only eligible big bubble events (with temporal links identity events)
    #event_lst, not_big_bubble_events = sieve_bbubbles_events(all_event_lst, "Reporting")
    #bubble_ids = {'T26','T28','T77','T31','T86','T32','T34','T33','T35'}

    excluded_events = set()
    for event_id in all_event_lst:
        #if event_id in bubble_ids:
            event_class = all_event_lst[event_id][0].get_attr_value("Class")
            event_text = join_tokens([tok.text for tok in all_event_lst[event_id]])

            msc_str += f"participant \"{event_text}\" as {event_id} {MAP_COLORS[event_class]}\n"
            excluded_events.add(event_id)

    rel_id_lst = set()
    for event_id in excluded_events:
        for rel in all_event_lst[event_id][0].relations:

            if rel.rel_id not in rel_id_lst and \
                    rel.rel_type.startswith("TLINK") and \
                    rel.toks[0].id_ann[0] in excluded_events:
                rel_name = rel.rel_type.split("_")[1]
                if rel.argn == 'arg2':
                    msc_str += f"{event_id} -> {rel.toks[0].id_ann[0]}: {rel_name}\n"
                else:
                    msc_str += f"{rel.toks[0].id_ann[0]} -> {event_id}: {rel_name}\n"
                rel_id_lst.add(rel.rel_id)

    msc_str += "@enduml"

    return msc_str
def build_fig_ann(ann_file, output_dir, lang="en"):

    url = "http://www.plantuml.com/plantuml/img/"
    pl = plantuml.PlantUML(url)

    reader = ReadBrat(lang)
    data = reader.process_file(ann_file)

    #only  get non-reporting events and the temporal relation links
    # between them
    msc_str = build_fig(data)
    ann_base_name = os.path.split(ann_file)[1]
    file_name_out = os.path.join(output_dir, f"{ann_base_name}_msc.txt")
    print("-->", file_name_out)

    with open(file_name_out, "w") as fd:
        fd.write(msc_str)

    file_name_fig = os.path.join(output_dir, f"{ann_base_name}_msc.png")
    print("-->", file_name_fig)
    success = pl.processes_file(file_name_out, outfile=file_name_fig)


import torch.nn as nn
import torch.nn.functional as F
from transformers import AutoModel, AutoTokenizer, PreTrainedModel
from .config import SRLModelConfig


class SRLModel(PreTrainedModel):
    config_class = SRLModelConfig

    def __init__(self, config):
        super().__init__(config)

        print(config.num_labels, config.bert_model_name, config.embedding_dropout)

        # Load pre-trained transformer-based model and tokenizer
        self.tokenizer = AutoTokenizer.from_pretrained(config.bert_model_name)
        self.transformer = AutoModel.from_pretrained(
            config.bert_model_name,
            num_labels=config.num_labels,
            output_hidden_states=True,
        )
        self.transformer.config.id2label = config.id2label
        self.transformer.config.label2id = config.label2id

        # The roberta models do not have token_type_embeddings
        # (the type_vocab_size is 1)
        # but we use this to pass the verb's position
        # so we need to change the model and initialize the embeddings randomly
        if "xlm" in config.bert_model_name or "roberta" in config.bert_model_name:
            self.transformer.config.type_vocab_size = 2
            # Create a new Embeddings layer, with 2 possible segments IDs instead of 1
            self.transformer.embeddings.token_type_embeddings = nn.Embedding(
                2, self.transformer.config.hidden_size
            )
            # Initialize it
            self.transformer.embeddings.token_type_embeddings.weight.data.normal_(
                mean=0.0, std=self.transformer.config.initializer_range
            )

        # Linear layer for tag projection
        self.tag_projection_layer = nn.Linear(
            self.transformer.config.hidden_size, config.num_labels
        )

        # Dropout layer for embeddings
        self.embedding_dropout = nn.Dropout(p=config.embedding_dropout)

        # Number of labels
        self.num_labels = config.num_labels

    def forward(self, input_ids, attention_mask, token_type_ids, labels=None):

        # print("FORWARD")
        # print(labels)

        # Forward pass through the transformer model
        # Returns BaseModelOutputWithPoolingAndCrossAttentions
        outputs = self.transformer(
            input_ids=input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
        )

        # Extract the [CLS] token representation
        # cls_output = outputs.pooler_output

        bert_embedding = outputs.last_hidden_state

        # Apply dropout to the embeddings
        embedded_text_input = self.embedding_dropout(bert_embedding)

        # Project to tag space
        logits = self.tag_projection_layer(embedded_text_input)

        reshaped_log_probs = logits.view(-1, self.num_labels)
        class_probabilities = F.softmax(reshaped_log_probs, dim=-1).view(
            logits.size(0), logits.size(1), -1
        )

        output_dict = {"logits": logits, "class_probabilities": class_probabilities}

        output_dict["attention_mask"] = attention_mask
        output_dict["input_ids"] = input_ids
        # output_dict["start_offsets"] = start_offsets

        if labels is not None:
            # print("Input", logits.view(-1, self.num_labels).size())
            # print("Target", labels.view(-1).size())
            # print("C", self.num_labels)
            # print("AllenNLP function", logits.size(-1))
            # Could consider passing ignore_index as 0 (pad index) for minor optimization
            loss = nn.CrossEntropyLoss(ignore_index=-100)(
                logits.view(-1, self.num_labels), labels.view(-1)
            )
            output_dict["loss"] = loss
        return output_dict



"""
    NLTK annotator

    Used for:
        - Actor extraction 
            'en' : default                
"""

from text2story.core.utils import chunknize_actors
from text2story.core.exceptions import InvalidLanguage

import nltk
from nltk import word_tokenize, sent_tokenize, pos_tag, ne_chunk, tree2conlltags


def load(lang):
    """
    Used, at start, to load the pipeline for the supported languages.
    """
    if lang == "en":
        nltk.download("maxent_ne_chunker")
        nltk.download('maxent_ne_chunker_tab')

        nltk.download('averaged_perceptron_tagger_eng')
        nltk.download('averaged_perceptron_tagger')
        nltk.download("words")


def extract_participants(lang, text):
    """
    Parameters
    ----------
    lang : str
        the language of text to be annotated
    text : str
        the text to be annotated
    
    Returns
    -------
    list[tuple[tuple[int, int], str, str]]
        the list of actors identified where each actor is represented by a tuple
    
    Raises
    ------
    InvalidLanguage if the language given is invalid/unsupported
    """

    if lang not in ['en']:
        raise InvalidLanguage(lang)

    language_mapping = {'en' : 'english'}

    iob_token_list = []

    char_offset = 0 
    sents = sent_tokenize(text, language=language_mapping[lang])
    
    for sent in sents:
        tree = ne_chunk(pos_tag(word_tokenize(sent, language=language_mapping[lang])))
        
        doc = tree2conlltags(tree) # doc :: [(Token, POS_TAG, IOB-NE)]
        
        for token in doc:
            token_text = token[0] # Don't call token[0] 'text', it will ofuscate the 'text' parameter!
            char_offset = text.find(token_text, char_offset) # always actualize char_offset to where we are, so next search we are doing, we are following the part of the text we didn't search yet
            char_span = (char_offset, char_offset + len(token_text))
            char_offset += len(token_text)
            pos = normalize(token[1])
            ne = token[2][:2] + normalize(token[2][2:]) if token[2] != 'O' else 'O'
            
            iob_token_list.append((char_span, pos, ne))
    
    actor_list = chunknize_actors(iob_token_list)
        
    return actor_list


def normalize(label):
    """
    Parameters
    ----------
    label : str
    
    Returns
    -------
    str
        the label normalized
    """

    mapping = {
        # POS tags (Penn Treebank Project: https://www.ling.upenn.edu/courses/Fall_2003/ling001/penn_treebank_pos.html)
        'NN'    : 'Noun',
        'NNS'   : 'Noun',
        'NNP'   : 'Noun',
        'NNPS'  : 'Noun',
        'PRP'   : 'Pronoun',
        'PRP$'  : 'Pronoun',
        'WP'    : 'Pronoun',
        'WP$'   : 'Pronoun',

        # NE labels
        'DATE'        : 'Date',
        'FACILITY'    : 'Loc',
        'GPE'         : 'Other',
        'GSP'         : 'Other', 
        'LOCATION'    : 'Loc',
        'MONEY'       : 'Other',
        'ORGANIZATION': 'Org',
        'PERCENT'     : 'Other',
        'PERSON'      : 'Per'
    }

    return mapping.get(label, 'UNDEF')



import platform

from text2story.core.exceptions import InvalidLanguage
import nltk
import joblib
from pathlib import Path

import os
import sys


from text2story.core.utils import normalize_tag, install

try:
    import torch
    from torchcrf import CRF
except ModuleNotFoundError:
    print("Warning: If you are going to use BERT NER PT module, you should install transformers, torch  and pytorch-crf libraries.")


pipeline = {}




def load(lang):
    if lang == "pt":
        try:
            import transformers
            import torch
            from torchcrf import CRF
        except ModuleNotFoundError:
            raise Exception("It is not possible to load BERT NER PT module, since some of \\"
                   "the required libraries are not installed: transformers,  torch or pytorch-crf")


        pipeline['pt']  = transformers.pipeline(
            model="arubenruben/NER-PT-BERT-CRF-Conll2003",
            device=torch.device("cuda:0") if torch.cuda.is_available() else torch.device("cpu"),
            trust_remote_code=True
        )
    else:
        raise InvalidLanguage(lang)


def extract_participants(lang, text):
    """
    Parameters
    ----------
    lang : str
        the language of text to be annotated
    text : str
        the text to be annotated

    Returns
    -------
    list[tuple[tuple[int, int], str, str]]
        the list of actors identified where each actor is represented by a tuple

    Raises
    ------
        InvalidLanguage if the language given is invalid/unsupported
    """
    if lang == "pt":
        return extract_actors_pt(text)
    else:
        raise InvalidLanguage(lang)

def extract_actors_pt(text):
    tokens = nltk.wordpunct_tokenize(text)
    result = pipeline['pt']({"tokens":tokens})

    # pos tagger: https://github.com/inoueMashuu/POS-tagger-portuguese-nltk
    current_path = Path(__file__).parent
    tagger = joblib.load(os.path.join(current_path, 'POS_tagger_brill.pkl'))

    pos_tags = tagger.tag(tokens)

    # find the tokens positions
    current_pos = 0
    pos_lst = []
    for tok in tokens:
        current_pos = text.find(tok, current_pos)
        pos_lst.append(current_pos)

    actor_lst = []
    i = 0
    last_label = 'O'
    current_actor = []
    while i < len(result):
        if result[i] != 'O':
            if last_label == 'O' and current_actor != []:
                actor_lst.append(current_actor)
                current_actor = []

            current_actor.append(i)
        last_label = result[i]
        i += 1

    if current_actor != []:
        actor_lst.append(current_actor)
    # the answer of the method
    ans = []
    for actor in actor_lst:
        idx_fst = actor[0]
        idx_last = actor[-1]

        tok_last = tokens[idx_last]

        position_fst = pos_lst[idx_fst]
        position_last = pos_lst[idx_last] + len(tok_last)

        start = position_fst
        end = len(text[position_fst:position_last])

        category = result[idx_fst].split('-')[1]

        normalized_postag = normalize_tag(pos_tags[idx_fst][1])
        ans.append(((start, end), normalized_postag , normalize_tag(category)))


    return ans




import text2story as t2s
from text2story.readers.read_brat import ReadBrat

from text2story.core.exceptions import InvalidIDAnn

def search_entity_span(entity, entity_lst, visited_entities):

    for entity_id in entity_lst:
        if entity.text == entity_lst[entity_id].text and\
            entity_id not in visited_entities:
            visited_entities.add(entity_id)
            return entity_id

def get_mapped_id(entity_id, entities_map):

    for entity_type in entities_map:
        if entity_id in entities_map[entity_type]:
            return entities_map[entity_type][entity_id]

def intersect_entities(doc1, doc2, entity_name):
    """

    @param doc1 Narrative:
    @param doc2 Narrative:
    @return: [(string, string)] A list of ids that doc1 and doc2 have in common.
    None is return if there is no element in common
    """

    entity_name_lst_1 = doc1.get_entity_byname(entity_name)
    entity_name_lst_2 = doc2.get_entity_byname(entity_name)

    visited_entities = set()

    intersection = {}

    for entity_id_1 in entity_name_lst_1:

        e1 = entity_name_lst_1[entity_id_1]
        entity_id_2 = search_entity_span(e1, entity_name_lst_2, visited_entities)

        if entity_id_2 is not None:
            intersection[entity_id_1] = entity_id_2

    return intersection

def is_same_relation_type(relation, relations_lst):

    for rel in relations_lst:
        if isinstance(rel, type(relation)) and rel.type == relation.type:
            return rel.id

def intersect_relations(doc1,doc2, intersections):

    intersection_rel = {}

    for link_type in doc1.links:
        for relation in doc1.links[link_type]:

            arg1_doc2 = get_mapped_id(relation.arg1, intersections)
            arg2_doc2 = get_mapped_id(relation.arg2, intersections)

            if arg1_doc2 is None:
                raise InvalidIDAnn(relation.arg1)

            if arg2_doc2 is None:
                raise InvalidIDAnn(relation.arg2)

            relations_lst = doc2.get_relations_byargs(arg1_doc2, arg2_doc2)
            relation_doc2_id = is_same_relation_type(relation, relations_lst)
            if relation_doc2_id is not None:
                intersection_rel[relation.id] = relation_doc2_id
            else:
                print("-->", relation.id)

    return intersection_rel

def compare_files(lang, file1, file2):

    reader = ReadBrat()

    token_lst_1 = reader.process_file(file1)
    token_lst_2 = reader.process_file(file2)

    doc1 = t2s.Narrative.fromTokenCorpus(lang, token_lst_1)
    doc2 = t2s.Narrative.fromTokenCorpus(lang, token_lst_2)

    intersections = {}

    intersections["participants"] = intersect_entities(doc1, doc2, "participants")
    intersections["events"] = intersect_entities(doc1, doc2, "events")
    intersections["times"] = intersect_entities(doc1, doc2, "times")
    intersections["spatial"] = intersect_entities(doc1, doc2, "spatial_relations")

    agreements = {}

    if len(doc1.participants) > 0:
        agreements["participants"] =  len(intersections["participants"]) / len(doc1.participants)
    else:
        agreements["participants"] = 0

    if len(doc1.events):
        agreements["events"] = len(intersections["events"]) / len(doc1.events)
    else:
        agreements["events"] = 0

    if len(doc1.times) > 0:
        agreements["times"] = len(intersections["times"]) / len(doc1.times)
    else:
        agreements["times"] = 0

    if len(doc1.spatial_relations) > 0:
        agreements["spatial"] = len(intersections["spatial"]) / len(doc1.spatial_relations)
    else:
        agreements["spatial"] = 0

    intersections_rel = intersect_relations(doc1, doc2, intersections)
    if len(doc1.links) > 0:
        total_links = sum([len(doc1.links[link_type]) for link_type in doc1.links])
        agreements["relations"] = len(intersections_rel) / total_links
    else:
        agreements["relations"] = 0

    return agreements


import os

from pyvis.network import Network

import text2story as t2s
from brat2drs import brat2drs
from drs2viz import parser

BRAT_ANN_DIR = './files'


def write_ann_to_file(iso_notation):
    f = open(f"{BRAT_ANN_DIR}/example.ann", "w")
    for line in enumerate(iso_notation):
        f.write(f'{line[1]}\n')
    f.flush()
    f.close()


def write_txt_to_file(text):
    f = open(f"{BRAT_ANN_DIR}/example.txt", "w")
    f.write(text)
    f.flush()
    f.close()


class Text2Viz:
    def __init__(self, language, text, publication_time, list_visualizations=None):
        self.narrative = t2s.Narrative(language, text, publication_time)
        self.models_loaded = False

        if list_visualizations is None:
            self.list_visualizations = ['ann_text', 'drs', 'graph']
        else:
            self.list_visualizations = list_visualizations

    def run(self):
        write_txt_to_file(self.narrative.text)

        if not self.models_loaded:
            print("Loading Models.")

            t2s.start()

            print("Finish Loading Models")

            self.models_loaded = True

        print("Done")

        iso_notation = self.sanitize_iso(self.extract_elements())

        write_ann_to_file(iso_notation)

        self.brat_2_drs(iso_notation)

        self.drs_2_viz()

    def extract_elements(self) -> str:

        if self.narrative.lang == 'en':
            self.narrative.extract_actors('spacy')
            self.narrative.extract_times('py_heideltime')
            self.narrative.extract_events('allennlp')
            self.narrative.extract_objectal_links('allennlp')
            self.narrative.extract_semantic_role_links('allennlp')

        elif self.narrative.lang == 'pt':
            self.narrative.extract_actors('spacy')
            self.narrative.extract_times('py_heideltime')
            self.narrative.extract_events('custompt')
            # self.narrative.extract_objectal_links('allennlp')
            # self.narrative.extract_semantic_role_links('allennlp')

        return self.narrative.ISO_annotation()

    def brat_2_drs(self, list_file_content: list):
        f_parser = brat2drs.file_parser(list_file_content)

        dexpr_list, f_parser = brat2drs.assign_variable(f_parser)
        dr_set, dexpr_list = brat2drs.attributes_events(dexpr_list, f_parser)
        dexpr_list = brat2drs.event_event_relation(f_parser, dexpr_list)
        actors, actors_sr = brat2drs.actors_relation(f_parser)

        relations, dexpr_list = brat2drs.TE_relations(f_parser, dexpr_list)
        relations, dexpr_list = brat2drs.TT_relations(relations, f_parser, dexpr_list)

        actors = brat2drs.update_actors(relations, actors, f_parser)

        output_file = f'{BRAT_ANN_DIR}/example_drs.txt'

        brat2drs.write_output(dexpr_list, dr_set, actors, relations, output_file)

    def drs_2_viz_graph(self, filename, notebook_bool=False):

        actors, non_ev_rels, ev_rels = parser.get_graph_data(f'./files/{filename}_drs.txt')

        net = Network(notebook=notebook_bool)

        for key in actors.keys():
            net.add_node(key, label=actors[key])

        for elem in non_ev_rels:
            net.add_edge(elem[0], elem[2], title=elem[1])

        for elem in ev_rels:
            net.add_edge(elem[0], elem[2], title=elem[1])

        net.show('mygraph.html')

    def drs_2_viz(self, notebook_bool=False):
        drs_file = parser.get_drs_files(BRAT_ANN_DIR)[0]
        filename = drs_file.split('/')[-1].split('_drs')[0]

        if 'ann_text' in self.list_visualizations:
            self.drs_2_viz_ann_text(filename)

        if 'drs' in self.list_visualizations:
            self.drs_2_viz_drs_txt(filename)

        # self.drs_2_viz_msc(filename)

        if 'graph' in self.list_visualizations:
            self.drs_2_viz_graph(filename, notebook_bool)

    def drs_2_viz_ann_text(self, filename):
        ann_file = os.path.join(BRAT_ANN_DIR, f'{filename}.ann')

        with open(ann_file, 'r') as fp:
            ann_text = fp.readlines()

        print(f"ANN_TEXT:\n-----------------------")

        for line in ann_text:
            print(line)

    def drs_2_viz_drs_txt(self, filename):
        with open(f'./files/{filename}_drs.txt', 'r') as fp:
            drs_text = fp.readlines()

        print(f"DRS_TEXT:-----------------------\n")

        for line in drs_text:
            print(line)

    def drs_2_viz_msc(self, drs_file):

        parser.get_msc_data(f'./files/{drs_file}_drs.txt')

    def sanitize_iso(self, raw_iso: str):
        list_elem = raw_iso.split('\n')

        parsed_list = [x.strip().replace('\t', ' ') for x in list_elem]

        return list(filter(lambda elem: elem != "", parsed_list))


def require_input_from_user():
    language = input(f"Enter language 'pt' for Portuguese or 'en' for English\n")

    if language != 'pt' and language != 'en':
        print("Not a valid language. Assuming English")
        language = 'en'

    text = input("Insert Text\n")

    return [language, text]


if __name__ == '__main__':
    text2viz = Text2Viz('en', '', '2021-12-31')

    while True:

        [language, text] = require_input_from_user()

        if language == 'pt':
            text2viz.narrative.lang = 'pt'
        elif language == 'en':
            text2viz.narrative.lang = 'en'

        text2viz.narrative.text = text

        text2viz.run()



import os
import pandas as pd
import shutil
from pathlib import Path

import string

# adding root directory
import sys
sys.path.insert(0, '..')
sys.path.insert(0, '../../')

from sklearn.pipeline import FeatureUnion
from sklearn.pipeline import Pipeline
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.preprocessing import OneHotEncoder
from sklearn.compose import ColumnTransformer
from sklearn.linear_model import LogisticRegression
from sklearn import metrics

from sklearn_crfsuite.utils import flatten
import sklearn_crfsuite
from sklearn_crfsuite import scorers
#from sklearn_crfsuite import metrics

from seqeval.metrics import classification_report
from seqeval.scheme import IOB2
import nlpnet


import joblib

from readers.read_brat import ReadBrat
from readers.token_corpus import TokenCorpus


class TokSRL:
    def __init__(self):
        self.gov_verb = None
        self.arg = None
        self.tok = None

class FeatureExtractor:

    def __init__(self):

        self.srl_tagger = nlpnet.SRLTagger("srl-pt/",language="pt")
        # (name, transformer, columns)
        self.preprocessor = ColumnTransformer(
                transformers=[ 
                    ('token', TfidfVectorizer(),"token"), #TfidfVectorizer accepts column name only between quotes
                    ('lemma', TfidfVectorizer(),"lemma"), #TfidfVectorizer accepts column name only between quotes
                    ('head', TfidfVectorizer(),"head"), #TfidfVectorizer accepts column name only between quotes
                    ('head_lemma', TfidfVectorizer(),"head_lemma"), #TfidfVectorizer accepts column name only between quotes
                    ('gov_verb', TfidfVectorizer(),"gov_verb"), #TfidfVectorizer accepts column name only between quotes
                    ('category', OneHotEncoder(), ["head_pos","pos"]),
                    ],
                    remainder="drop"
                )

        self.pipe = Pipeline(
                steps=[
                    ('preprocessor', self.preprocessor),
                    ('classifier', LogisticRegression()),
                    ],
                )

    def get_verb_pos(self, verb, tok_lst, idx_start):
        """
        It get the verb index in a given token list. The first occurrence,
        starting from idx_start index position

        @param string: the verb
        @param [string]: a list of tokens 
        @param int: the starting point to start the search for the verb

        @return int: the verb index in the token list
        """

        idx = idx_start
        while idx < len(tok_lst):
            if tok_lst[idx] == verb:
                break
            idx = idx + 1

        return idx

    def set_role_token(self, tokcorpus_lst, idx_verb, args):
        """
        It sets the role position for each token in the tokcorpus_lst list

        @param [TokenCorpus]: a list of TokenCorpus objects
        @param int: the index of the current governing verb of the given argument
                  structure
        @param dict: The dictionary that stores the argument structure of an SRL 
                    labeling returned by the nlpnet API

        @return [TokenCorpus]: the list of TokenCorpus objects updated with 
                    governing verb and srl role
        """

        # TODO 1: It is possible implement a more efficient method?
        # TODO 2: Perform a more complete test to see what examples does not work
        # here

        for role in args:
            for tok in args[role]:
                # talvez começar analise APOS idx_verb e, tokcorpus_lst!
                # nao pode, pois tem argumentos antes de idx_verb
                # Verbo mais proximo? Como saber qual eh o mais proximo?
                #  guardar posicao do gov_verb
                for idx_corpus, tokcorpus in enumerate(tokcorpus_lst):
                    if tok == tokcorpus.text:

                        # if still there is no role
                        if tokcorpus.srl is None:
                            tokcorpus.gov_verb = tokcorpus_lst[idx_verb].text
                            tokcorpus.gov_verb_idx = idx_verb
                            tokcorpus.srl = role
                            break
                        else:
                            # if there is role, but there is a more specific governing verb
                            # then assign a new role and a new governing verb
                            if tokcorpus.gov_verb_idx is not None:
                                # the nearest verb is more probable to be the gpverning verb
                                # so, this kind a heuristic according to the 
                                # nlpnet output
                                dist_verb = abs(idx_corpus - idx_verb)
                                dist_gov_verb = abs(idx_corpus - tokcorpus.gov_verb_idx)
                                if tokcorpus.gov_verb != tokcorpus_lst[idx_verb].text and\
                                        dist_verb <= dist_gov_verb:
                                    tokcorpus.gov_verb = tokcorpus_lst[idx_verb].text
                                    tokcorpus.srl = role

        #for tokcorpus in tokcorpus_lst:
        #    print(tokcorpus.text, tokcorpus.srl, tokcorpus.gov_verb)
        return tokcorpus_lst

    def get_srl_tags(self, doc):


        tok_lst = [tok.text for tok in doc]
        tags = self.srl_tagger.tag_tokens(tok_lst)

        all_toks_lst = []

        # tracking position of the verbs in the argument 
        # structures
        last_verb_idx = 0
        new_arg_lst = []
        for (verb, args) in tags.arg_structures:
            idx_verb = self.get_verb_pos(verb, tags.tokens, last_verb_idx)
            last_verb_idx = idx_verb
            new_arg_lst.append((verb, args, idx_verb))

        # mapping roles to tokens, using the governing verb positio
        tokcorpus_lst = [TokenCorpus(tok) for tok in tags.tokens]
        for verb, args, idx_verb in new_arg_lst:
            tokcorpus_lst = self.set_role_token(tokcorpus_lst, idx_verb, args)
        all_toks_lst = all_toks_lst + tokcorpus_lst
        return all_toks_lst

    def doc2feature(self, doc, file_name):

        data_= {"token":[],"lemma":[],"head":[],"head_lemma":[],\
                "head_pos":[],"pos":[]}

        y_ = []

        srl_tags = self.get_srl_tags(doc)

        for tok in doc:

            data_["token"] = tok.text
            data_["lemma"] = tok.lemma
            data_["head"] = tok.head
            data_["head_lemma"] = tok.head_lemma
            data_["gov_verb"] = tok.gov_verb

            data_["head_pos"] = tok.head_pos
            data_["pos"] = tok.pos

            if tok.ann == "Event":
                y_.append("I")
            else:
                y_.append("O")

        return data_,y_


    def doc2featureCRF(self, doc, file_name):

        data_doc = []
        y_doc = []

        sent_lst = []
        sent = []
        for tok in doc:
            if  tok.text != ".":
                sent.append(tok)
            else:
                sent.append(tok)
                sent_lst.append(sent)
                sent = []

        if sent != []:
            sent_lst.append(sent)

        srl_toks = []
        for sent in sent_lst:
            srl_toks += self.get_srl_tags(sent)


        last_label = "O"
        for idx, tok in enumerate(doc):
            
            data_= {}

            data_["token"] = tok.text
            data_["lemma"] = tok.lemma
            data_["head"] = tok.head
            data_["head_lemma"] = tok.head_lemma
            #data_["gov_verb"] = tok.gov_verb

            data_["head_pos"] = tok.head_pos
            data_["pos"] = tok.pos

            if tok.ann == "Event" and \
                    (last_label == "I-EVENT" or last_label == "B-EVENT"):
                y_doc.append("I-EVENT")
                last_label = "I-EVENT"
            elif tok.ann == "Event" and last_label == "O":
                y_doc.append("B-EVENT")
                last_label = "B-EVENT"
            else:
                y_doc.append("O")
                last_label = "O"

            data_doc.append(data_)

        return data_doc, y_doc

    def _get_events_iob(self, doc, class_tokens):
        """
        Get a list of events from the labels of a iob labeling list
        """

        last_class = "O"
        count_event = 1
        current_event = {}
        event_lst = []

        symbols = string.punctuation + "\”\"\“"

        # get events

        for idx, tok in enumerate(doc):

            if class_tokens[idx] == "B-EVENT":

                current_event = {"id":count_event, "start":tok.offset, \
                        "end":tok.offset + len(tok.text),\
                        "value":tok.text}

                last_class = "B-EVENT"
                count_event += 1

                #print("T%d Event %d %d %s" % \
                #        (count_event, tok.offset, tok.offset + len(tok.text), tok.text))
            elif class_tokens[idx] == "I-EVENT":

                if last_class == "B-EVENT" or last_class == "I-EVENT":

                    id_ = current_event["id"]
                    start_ = current_event["start"]
                    end_ = tok.offset + len(tok.text)
                    if tok.text not in symbols:
                        value_ = current_event["value"] + " " + tok.text
                    else:
                        value_ = current_event["value"] + tok.text

                    current_event = {"id":id_,"start":start_,"end":end_, "value":value_}
                    last_class = "I-EVENT"
                else:
                    current_event = {"id":count_event, "start":tok.offset, \
                        "end":tok.offset + len(tok.text),\
                        "value":tok.text}

                    last_class = "I-EVENT"
                    count_event += 1
            else:
                if last_class == "B-EVENT" or last_class == "I-EVENT":
                    event_lst.append(current_event)
                    current_event = {}
                    last_class = "O"

        return event_lst

    def write2brat_file(self, doc, class_tokens, output_file):

        event_lst = self._get_events_iob(doc, class_tokens)

        with open(output_file, "w") as fd_output:
            for event in event_lst:
                fd_output.write("T%d\tEvent %d %d\t%s\n" % \
                        (event["id"],\
                        event["start"],\
                        event["end"],\
                        event["value"]))


    def write2brat(self, data_tokens, file_txt_lst, y_pred):

        # TODO: preciso do offset aqui tbm
        # coloquei offset.

        if not(os.path.exists("brat_output")):
            os.mkdir("brat_output")

        for idx, doc in enumerate(data_tokens):

            # copy the txt file
            shutil.copy2(file_txt_lst[idx], "brat_output")

            # write the annotation file
            ann_file = os.path.join("brat_output",Path(file_txt_lst[idx]).stem + ".ann")


            self.write2brat_file(doc, y_pred[idx], ann_file)

        

    def process_test(self, data_dir,model="crf",write_output=False):

        reader = ReadBrat()
        data_tokens = reader.process(data_dir)

        data_lst = []
        data_dct = {}
        y = []
        idx = 0

        for doc in data_tokens:
            if model == "crf":
                data_doc, y_doc = self.doc2featureCRF(doc, reader.file_lst[idx])
                data_lst.append(data_doc)
                y.append(y_doc)
            else:
                data_doc, y_doc = self.doc2feature(doc, reader.file_lst[idx])
                data_dct = {**data_doc, **data_dct}
            idx += 1

        clf = joblib.load("crf.joblib")
        y_pred = clf.predict(data_lst)

        print(metrics.classification_report(flatten(y), flatten(y_pred)))

        if write_output:
            self.write2brat(data_tokens, reader.file_lst, y_pred)

    def process(self, data_dir, model="crf"):
        """
        Read a data_dir (.ann and their .txt files)
        and returns a matrix of features, and labels

        @param string: data directory

        @return (numpy.array,numpy.array): Matrix of features and vector 
        of labels
        """

        # read file names .ann and .txt
        reader = ReadBrat()
        data_tokens = reader.process(data_dir)

        data_lst = []
        data_dct = {}
        y = []
        idx = 0
        for doc in data_tokens:
            if model == "crf":
                data_doc, y_doc = self.doc2featureCRF(doc, reader.file_lst[idx])
                data_lst.append(data_doc)
                y.append(y_doc)
            else:
                data_doc, y_doc = self.doc2feature(doc, reader.file_lst[idx])
                data_dct = {**data_doc, **data_dct}
            idx += 1


        if model != "crf":
            df = pd.DataFrame(data_dct)
            if os.path.exists("preprocessor.joblib"):
                print("Warning: Loading a TF-IDF local model called preprocessor.joblib.")
                self.preprocessor = joblib.load("preprocessor.joblib")
            else:
                print("Warning: Creating a TF-IDF new model from the given data.")
                self.preprocessor.fit(df)
                joblib.dump(self.preprocessor,"preprocessor.joblib")

            df_trans = self.preprocessor.transform(df)
            # TODO: finish model
        else:
            clf = sklearn_crfsuite.CRF(algorithm='lbfgs',
                    c1=0.1, c2=0.1, max_iterations=100, 
                    all_possible_transitions=True)
            df_trans = data_lst

        clf.fit(df_trans, y)
        joblib.dump(clf, "crf.joblib")

if __name__ == "__main__":

    data_dir = "data/train/"
    test_dir = "data/test/"
    ftr_extractor = FeatureExtractor()

    ftr_extractor.process(data_dir)
    ftr_extractor.process_test(test_dir, write_output=True)




import os
from pathlib import Path

from bs4 import BeautifulSoup

from text2story.readers import token_corpus
from text2story.readers.read import Read
import xml.etree.ElementTree as ET

import spacy
from text2story.readers.token_corpus import TokenRelation
import re
from text2story.core.utils import find_substring_match

class ReadACE(Read):

    def __init__(self):
        self.file_lst = []
        if not(spacy.util.is_package('en_core_web_lg')):
            spacy.cli.download('en_core_web_lg')

    def _get_rel_mention_data(mentions, raw_text):

        info_lst = []
        current_idx = 0
        mention_lst =  []

        for mention in mentions:
            mention_id = mention.get('ID')
            #print("-->", mention_id) CNN_CF_20030303.1900.00-E84-167
            arguments_lst = mention.findall(".//relation_mention_argument")
            args = []
            for arg in arguments_lst:
                extent = arg.find(".//extent/charseq")
                arg_id = arg.get("REFID")
                arg_role = arg.get("ROLE")
                args.append((arg_id, arg_role))

            mention_lst.append((mention_id, args))

        return mention_lst


    @staticmethod
    def _get_mention_data(mentions, raw_text, mention_regex=".//extent/charseq"):

        info_lst = []
        current_idx = 0
        mention_lst =  []

        for mention in mentions:
            mention_id = mention.get('ID')
            #print("-->", mention_id) CNN_CF_20030303.1900.00-E84-167

            extent = mention.find(mention_regex)
            start_mention = extent.get('START')
            end_mention = extent.get('END')

            mention_lst.append((mention_id, extent.text, int(start_mention), int(end_mention)))

        mention_lst.sort(key = lambda x:x[2])
        old_mention_start = None
        old_mention_end = None

        for mention in mention_lst:

            mention_id, mention_text, mention_start, mention_end = mention

            if old_mention_start !=None and old_mention_end !=None:
                if mention_start>= old_mention_start and mention_end<=old_mention_end:
                    # if the current mention is inside of the mention before
                    # the current_idx should be restarted
                    current_idx = start

            pattern = r'\b' + re.escape(mention_text) + r'\b'
            matches = re.finditer(pattern, raw_text, flags=re.IGNORECASE)
            span_idx = find_substring_match(matches, current_idx)

            if span_idx is None:
               span_idx = raw_text.find(mention_text, current_idx)

            start = span_idx
            end = span_idx + len(mention_text)

            current_idx = end
            old_mention_start = mention_start
            old_mention_end = mention_end


            info_lst.append((mention_id, mention_text, start, end))

        return info_lst

    @staticmethod
    def _get_participant_annotations(root, raw_text):

        participant_lst = []  # first the participants

        for entity in root.findall(".//entity"):
            entity_type = entity.get('TYPE')
            entity_mentions = entity.findall('entity_mention')

            mention_lst = ReadACE._get_mention_data(entity_mentions, raw_text)

            participant_lst.append(({'Participant_Type_Domain':entity_type}, mention_lst))

        return participant_lst

    @staticmethod
    def _map2isorelation(ace_relation, ace_subrelation):
        ace_relation2iso_relation = {
            "PART-WHOLE":"OLINK_partOf",
            "ORG-AFF":"OLINK_memberOf",
            "PHYS":"QSLINK_ground"
        }

        return ace_relation2iso_relation.get(ace_relation)
    @staticmethod
    def _map2isocategory(ace_category):

        # mapping generated by chatGPT
        event_type_to_category = {
            "Conflict": "Occurrence",
            "Contact": "Contact",
            "Transaction": "Transaction",
            "Movement": "Occurrence",
            "Personnel": "State",
            "Life": "Occurrence",
            "Business": "State",
            "Justice": "State",
            "Contact.Meet": "Reporting",
            "Contact.Correspondence": "Reporting",
            "Contact.Broadcast": "Reporting",
        }

        # Adding additional event types
        #event_type_to_category["Contact.Phone-Write"] = "Reporting"
        #event_type_to_category["Contact.Speech"] = "Perception"
        #event_type_to_category["Transaction.Transfer-Money"] = "Transaction"
        #event_type_to_category["Business.Merge"] = "Business"
        #event_type_to_category["Business.Declare-Bankruptcy"] = "Business"
        #event_type_to_category["Life.Marry"] = "State"

        return event_type_to_category[ace_category]

    @staticmethod
    def _get_relations_annotations(root, raw_text):
        relations_lst = []

        for relation in root.findall(".//relation"):
            relation_type = relation.get('TYPE')
            relation_subtype = relation.get('SUBTYPE')
            relation_mentions = relation.findall(".//relation_mention")
            mention_lst = ReadACE._get_rel_mention_data(relation_mentions, raw_text)

            relations_lst.append(({'Relation_Type': relation_type,'Relation_Subtype': relation_subtype}, mention_lst))

        return relations_lst

    @staticmethod
    def _get_event_annotations(root, raw_text):

        event_lst = []  # then the events

        for event in root.findall(".//event"):
            event_type = event.get('TYPE')
            event_subtype = event.get('SUBTYPE')
            modality = event.get('MODALITY')
            event_mentions = event.findall(".//event_mention")

            mention_lst = ReadACE._get_mention_data(event_mentions, raw_text, mention_regex=".//anchor/charseq")

            event_lst.append(({'Class':ReadACE._map2isocategory(event_type), \
                               'Subclass':event_subtype}, mention_lst))


        return event_lst

    @staticmethod
    def _get_timex_annotations(root, raw_text):

        timex_lst = []  # then the events
        for timex in root.findall(".//timex2"):

            timex_id = timex.get('ID')
            timex_val = timex.get('VAL')
            timex_anchor_val = timex.get('ANCHOR_VAL')
            timex_anchor_dir = timex.get('ANCHOR_DIR')

            timex_mentions = timex.findall(".//timex2_mention")

            mention_lst = ReadACE._get_mention_data(timex_mentions, raw_text)
            timex_lst.append(({'Timex_Id':timex_id, 'Timex_Val':timex_val,\
                                'Timex_Anchor_Val':timex_anchor_val, 'Timex_Anchor_Dir':timex_anchor_dir}, mention_lst))

        return timex_lst

    def _read_label_file(self, apf_file_path, raw_text):

        # Load and parse the APF XML file
        tree = ET.parse(apf_file_path)
        root = tree.getroot()

        current_idx = 0

        # Iterate through the APF XML to extract annotations
        participant_lst = self._get_participant_annotations(root, raw_text)  # first the participants

        event_lst = self._get_event_annotations(root, raw_text)  # then the events

        timex_lst = self._get_timex_annotations(root, raw_text)

        relations_lst = self._get_relations_annotations(root, raw_text)

        return participant_lst, event_lst, timex_lst, relations_lst

    @staticmethod
    def _search_subtoken(start_idx, end_idx, annotations):

        ann_lst = []
        for category in annotations:

            for label_lst in annotations[category]:
                if label_lst == []:
                    continue

                mention_lst = label_lst[-1]

                for  mention in mention_lst:
                    mention_id, mention_text, mention_start, mention_end = mention

                    # the given token is inside this mention
                    if  start_idx >= mention_start and end_idx <= mention_end:
                        mention_annotations = label_lst[0]
                        if mention_annotations != {}:
                            ann_lst.append((category, mention_annotations, mention))
        return ann_lst

    @staticmethod
    def _read_text_file(data_file):
        with open(data_file, "r") as fd:
            file_content = fd.read()
            soup = BeautifulSoup(file_content, 'html.parser')

        # 'bs4.element.ResultSet'
        body = soup.doc.body
        text = ""

        for child in body.children:
            if child.name == 'headline' or child.name == 'text':
                text = text + child.text

        return text.strip()

    def process_file(self, data_file, lang="en"):
        """
        Process an ACE file format (.sgm extension) and return a tokenCorpus list.

        It assumes that in the same path of .sgm file there is an .apf.xml file with the
        same name.
        @param data_file:
        @return:
        """

        label_file = os.path.splitext(data_file)[0] + ".apf.xml"
        text = ReadACE._read_text_file(data_file)

        # get the labels
        participant_lst, event_lst, timex_lst, relations_lst = self._read_label_file(label_file, text)

        annotations = {'Participant':participant_lst,\
                       'Event':event_lst,\
                       'Time':timex_lst}

        # tokenization of the raw text and serialization to a list of TokenCorpus data type
        # if I use spacy to tokenization, then I have idx
        nlp = spacy.load('en_core_web_lg')

        doc = nlp(text)
        tok_lst = []

        sent_id = -1
        ref_count = 1

        # a map between referecens in ACE dataset and new ids in TokenCorpus format
        mention2ref = {}
        for tok_count, tok in enumerate(doc):

            if tok.is_sent_start:
                sent_id += 1

            mytok = token_corpus.TokenCorpus()
            mytok.id = tok_count
            mytok.text = tok.text
            mytok.lemma = tok.lemma_
            mytok.pos = tok.pos_
            mytok.dep = tok.dep_
            mytok.head = tok.head.text
            mytok.head_pos = tok.head.pos_
            mytok.head_lemma = tok.head.lemma_
            mytok.offset = tok.idx
            mytok.sent_id = sent_id

            ans_sub = ReadACE._search_subtoken(tok.idx, tok.idx + len(tok.text),\
                                           annotations)
            for ans in ans_sub:
                category, ann, mention = ans
                mention_id, _ , _ , _ = mention

                # this mention already exists as TokenCorpus Annotation?
                if mention_id in mention2ref:
                    # yes!
                    ref_tok, _ = mention2ref[mention_id]
                    mytok.id_ann.append(ref_tok)
                else:
                    # otherwise create another ref
                    mytok.id_ann.append(f'T{ref_count}')
                    mention2ref[mention_id] = (f'T{ref_count}', mytok)
                    ref_count += 1

                mytok.attr.append((category, ann))

            tok_lst.append(mytok)

        # add relations that this token may be part of
        rel_count = 1
        for rel in relations_lst:

            attr, mention_lst = rel
            relation_type = attr['Relation_Type']
            relation_subtype = attr['Relation_Subtype']

            for mention in mention_lst:
                _, args = mention # args it is a list (arg_id, arg_role)

                tok_arg_lst = []
                not_mapped_flag = False
                for (arg_id, arg_role) in args:
                    if arg_id not in mention2ref:
                        print(f"Warning: {arg_id} is not a mapped element.")
                        not_mapped_flag = True
                        continue
                    ref_tok, tok = mention2ref[arg_id]
                    tok_arg_lst.append((tok, arg_role))

                if not_mapped_flag:
                    continue
                # now update each token in the arguments with the relation
                iso_relation = ReadACE._map2isorelation(relation_type, relation_subtype)
                # if the iso_relation is None, there is no possible map for the internal representation of an ISO relation
                if iso_relation is not None:
                    for tok, arg_role in tok_arg_lst:
                        arg_id = arg_role.lower().replace("-","") # in this case, arg_id is the number of argument arg1 or arg2
                        tok.relations.append(TokenRelation(f'R{rel_count}', [t for t, _ in tok_arg_lst], iso_relation, arg_id=arg_id))
                    rel_count += 1

        # now we can add the OLINK_objIdentity of the same entity mentions
        for mention_id1 in mention2ref:
            for mention_id2 in mention2ref:
                if mention_id1 != mention_id2:
                    stem_id1 = mention_id1.split('-')
                    stem_id1 = "-".join(stem_id1[:-1])



        return tok_lst

    def process(self, data_dir, output_dir=None, split=None):
        """
        It reads a set of files of annotations and text, then returns
        the processed tokens as TokenCorpus type

        @param string: path of data to gather and process
        @param string: if an output directory is provided, then the text files will be put there
        @param string: a file name that contains a list of the files in data dir that
        should be processed by this reader


        @return [[TokenCorpus]]: a list of lists of tokens
        """

        data_tokens = []

        for dirpath, dirnames, filenames in os.walk(data_dir):
            for f in filenames:
                if f.endswith(".sgm"):

                    p = Path(f)
                    data_file = os.path.join(dirpath, f)
                    print(f"Processing {data_file}")
                    token_lst = self.process_file(data_file)

                    if output_dir != None:
                        output_txt = os.path.join(output_dir, p.stem + ".txt")
                    else:
                        output_txt = os.path.join(dirpath, p.stem + ".txt")

                    text = ReadACE._read_text_file(data_file)
                    with open(output_txt, "w") as fd:
                        fd.write(text)


                    self.file_lst.append(output_txt)

                    if len(token_lst) > 0:
                        data_tokens.append(token_lst)
        return data_tokens




class TokenCorpus:
    """
    This class should store the token text, and
    annotation information
    """

    def __init__(self, text=None, token_id=None, sentence=None):
        # fields in the Token object
        self.text = text
        self.id = token_id
        self.lemma = None
        self.ann = None
        self.sentence = sentence
        self.pos = None
        self.dep = None
        self.head = None
        self.head_pos = None
        self.head_lemma = None
        self.gov_verb = None
        self.gov_verb_idx = None
        self.srl = None
        self.sent_id = None
        self.clause_id = None

        self.offset = None
        self.attr = []
        self.ann_offset = []

        # a token can be annotated by more than one 
        # layer, therefore it can present more than one 
        # id
        self.id_ann = []

        # a list of items of type TokenRelation
        self.relations = []

    def get_attr_value(self,attr_name, token_type=None):
        for (type_, attr_map) in self.attr:
            if token_type != None and type_ == token_type:
                return attr_map.get(attr_name)


    def is_type(self, token_type):
        for (type_, _) in self.attr:
            if type_ == token_type:
                return True
        return False

class TokenRelation:
    """
    Specifies a Relation between tokens
    """
    def __init__(self, rel_id, toks = [], rel_type = None, argn = None, arg_id = None):
        self.rel_id = rel_id
        self.toks = toks
        self.rel_type = rel_type
        self.argn = argn
        self.arg_id = arg_id



'''
    PY_HEIDELTIME annotator (https://github.com/JMendes1995/py_heideltime)

    Used for:
        - Timexs extraction
            'en' : default
            'pt' : default
'''
import traceback

#from text2story.core.exceptions import InvalidLanguage

from py_heideltime import py_heideltime
import re
import sys

def load(lang):
    """
    Used, at start, to load the pipeline for the supported languages.
    """
    pass # Nothing to load


def extract_times(lang, text, publication_time):
    """
    Parameters
    ----------
    lang : str
        the language of text to be annotated

    text : str
        the text to be annotated

    Returns
    -------
    list[tuple[tuple[int, int], str, str]]
        a list consisting of the times identified, where each time is represented by a tuple
        with the start and end character offset, it's value and type, respectively

    Raises
    ------
    InvalidLanguage if the language given is invalid/unsupported
    """

    if lang not in ['en', 'pt']:
        #raise InvalidLanguage
        print("Invalid Language")
        raise sys.exit(0)

    lang_mapping = {'pt' : 'Portuguese', 'en' : 'English'}
    lang = lang_mapping[lang]

    # annotations = py_heideltime(re.escape(text), language=lang, document_creation_time=publication_time)
    try:
        annotations = py_heideltime.heideltime(text, lang, "news")
    except IndexError as e:
        #traceback.print_exc()
        annotations = []
        print()
        print("WARNING: no time expression processed in this document for some bug in py_heideltime.")

    timexs = []
    for ann in annotations:
        start, end = ann["span"][0], ann["span"][1]
        timex = ((start,end), ann["type"], ann["text"])
        timexs.append(timex)

    return timexs


"""
    spaCy annotator

    Used for:
        - Actor extraction
            'pt' : https://spacy.io/models/en#en_core_web_lg)
            'en' : https://spacy.io/models/en#en_core_web_lg)
"""

from text2story.core.utils import chunknize_actors, normalize_tag
from text2story.core.exceptions import InvalidLanguage

import spacy

pipeline = {}

def load(lang):
    """
    Used, at start, to load the pipeline for the supported languages.
    """
    if lang == "pt":
        if not(spacy.util.is_package('pt_core_news_lg')):
            spacy.cli.download('pt_core_news_lg')
        pipeline['pt'] = spacy.load('pt_core_news_lg')

    elif lang == "en":
        if not(spacy.util.is_package('en_core_web_lg')):
            spacy.cli.download('en_core_web_lg')

        pipeline['en'] = spacy.load('en_core_web_lg')
    else:
        raise InvalidLanguage(lang)

    
def extract_participants(lang, text):
    """
    Parameters
    ----------
    lang : str
        the language of text to be annotated
    text : str
        the text to be annotated
    
    Returns
    -------
    list[tuple[tuple[int, int], str, str]]
        the list of actors identified where each actor is represented by a tuple

    Raises
    ------
        InvalidLanguage if the language given is invalid/unsupported
    """

    if lang not in ['pt', 'en']:
        raise InvalidLanguage(lang)

    doc = pipeline[lang](text)

    iob_token_list = []
    for token in doc:
        start_character_offset = token.idx
        end_character_offset = token.idx + len(token)
        character_span = (start_character_offset, end_character_offset)
        pos = normalize_tag(token.pos_)
        ne = token.ent_iob_ + "-" + normalize_tag(token.ent_type_) if token.ent_iob_ != 'O' else 'O'

        iob_token_list.append((character_span, pos, ne))

    actor_list = chunknize_actors(iob_token_list)

    return actor_list  





import importlib
import sys
import os
from pathlib import Path

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from text2story.core.exceptions import InvalidNarrativeComponent, DuplicateNarrativeComponent, InvalidTool
from typing import List
from types import ModuleType

from collections import ChainMap

sys.path.insert(0, str(Path(__file__).parent))

PARTICIPANT_EXTRACTION_TOOLS = {'spacy':['pt','en'], 'nltk':['en'], 'bertnerpt':['pt'], 'dbpedia':['pt','en'],'srl':['pt','en']}
#TIME_EXTRACTION_TOOLS = {'py_heideltime':['pt','en'], 'tei2go':['pt','en','it','de','es','fr']}
TIME_EXTRACTION_TOOLS = {'py_heideltime':['pt','en']}
EVENT_EXTRACTION_TOOLS = {"srl":["pt","en"]}
#OBJECTAL_LINKS_RESOLUTION_TOOLS = {'allennlp':['en']}
OBJECTAL_LINKS_RESOLUTION_TOOLS = {'maverick':['en']}
SEMANTIC_ROLE_LABELLING_TOOLS = {"srl":["pt","en"]}

LOCAL_ANNOTATORS = set()

def get_tools():
    return dict(ChainMap(PARTICIPANT_EXTRACTION_TOOLS, TIME_EXTRACTION_TOOLS, \
             EVENT_EXTRACTION_TOOLS, OBJECTAL_LINKS_RESOLUTION_TOOLS, \
                         SEMANTIC_ROLE_LABELLING_TOOLS))

def get_tool_list(tool_type:str):
    """

    @param tool_type: given a type of a tool , i.e, one of the following items ['participant','time', 'event','objectal_links',
                     'semantic_links']
    @return: the dictionary of tools names or None if it is not found
    """

    if tool_type == 'participant':
        return PARTICIPANT_EXTRACTION_TOOLS
    elif tool_type == 'event':
        return EVENT_EXTRACTION_TOOLS
    elif tool_type == 'time':
        return TIME_EXTRACTION_TOOLS
    elif tool_type == 'objectal_links':
        return OBJECTAL_LINKS_RESOLUTION_TOOLS
    elif tool_type == 'semantic_links':
        return  SEMANTIC_ROLE_LABELLING_TOOLS
    else:
        raise InvalidNarrativeComponent(tool_type)
def add_tool(tool_name:str, lang_lst:List[str], tool_types:List[str]):
    """
    It adds a tool to the gallery of annotators.

    @param tool: the name of the module where is your annotator
    @param lang_lst: a list of languages supported by your custom annotator,
                        for instance, ['en','fr']
    @param tool_types: a list of types that your annotator labels. The
                     supported types are ['participant','time', 'event','objectal_links',
                     'semantic_links']
    @return: None
    """

    LOCAL_ANNOTATORS.add(tool_name)

    for ttype in tool_types:
        tool_list = get_tool_list(ttype)
        if tool_name not in tool_list:
            tool_list[tool_name] = lang_lst
        else:
            raise DuplicateNarrativeComponent(tool_name)

def load(lang, tools=None):
    """

    It loads models associated with the tools aggregated in the text2story pipeline

    @param str lang: The language (pt, en, fr, de, it, es)
    @param [str] tools(optional): a list of the tools (options are: spacy, nltk, py_heideltime, allennlp, bertnerpt, tei2go)
    to be employed, if None is given, all availables tools for that language are loaded
    @return: None
    """

    available_tools = get_tools()
    if tools is None:
        tools = available_tools

    # loading the local annotators
    for tl in tools:
        if tl in LOCAL_ANNOTATORS and lang in tools[tl]:
            local_annotator = importlib.import_module(tl)
            local_annotator.load(lang)

    # the importing of libraries is made upon use
    # the user, thus, has to check the installation
    if lang in tools["spacy"] and "spacy" in tools:
        from text2story.annotators import SPACY
        SPACY.load(lang)
    if lang in tools["nltk"] and "nltk" in tools:
        from text2story.annotators import NLTK
        NLTK.load(lang)
    if lang in tools["py_heideltime"] and "py_heideltime" in tools:
        from text2story.annotators import PY_HEIDELTIME
        PY_HEIDELTIME.load(lang)
    if lang in tools["bertnerpt"] and "bertnerpt" in tools:
        from text2story.annotators import BERTNERPT
        BERTNERPT.load(lang)
    #if lang in tools["tei2go"]  and "tei2go" in tools:
    #    from text2story.annotators import TEI2GO
    #    TEI2GO.load(lang)
    if lang in tools["dbpedia"]  and "dbpedia" in tools:
        from text2story.annotators import DBPEDIA
        DBPEDIA.load(lang)
    if lang in tools["srl"]  and "srl" in tools:
        from text2story.annotators import SRL
        SRL.load(lang)
    if lang in tools["maverick"]  and "maverick" in tools:
        from text2story.annotators import MAVERICK
        MAVERICK.load(lang)

def get_available_tools():
    return PARTICIPANT_EXTRACTION_TOOLS.keys()+\
            TIME_EXTRACTION_TOOLS.keys()+\
            EVENT_EXTRACTION_TOOLS.keys()+\
            OBJECTAL_LINKS_RESOLUTION_TOOLS.keys()+\
            SEMANTIC_ROLE_LABELLING_TOOLS.keys()

def get_participant_tools():
    return PARTICIPANT_EXTRACTION_TOOLS

def get_time_tools():
    return TIME_EXTRACTION_TOOLS

def get_event_tools():
    return EVENT_EXTRACTION_TOOLS

def get_srlink_tools():
    return SEMANTIC_ROLE_LABELLING_TOOLS

def extract_participants(tool, lang, text, url=None):

    if tool in LOCAL_ANNOTATORS:
        local_annotator = importlib.import_module(tool)
        return local_annotator.extract_participants(lang, text)

    if tool == 'spacy':
        from text2story.annotators import SPACY
        return SPACY.extract_participants(lang, text)
    elif tool == 'nltk' and lang == "en":
        from text2story.annotators import NLTK
        return NLTK.extract_participants(lang, text)
    elif tool == 'bertnerpt':
        from text2story.annotators import BERTNERPT
        return BERTNERPT.extract_participants(lang, text)
    elif tool == 'dbpedia':
        from text2story.annotators import DBPEDIA
        return DBPEDIA.extract_participants(text, lang, url)
    if tool == 'srl':
        from text2story.annotators import SRL
        return SRL.extract_participants(lang, text)
    else:
        raise InvalidTool


def extract_times(tool, lang, text, publication_time):

    if tool in LOCAL_ANNOTATORS:
        local_annotator = importlib.import_module(tool)
        return local_annotator.extract_times(lang, text)

    if tool == 'py_heideltime':
        from text2story.annotators import PY_HEIDELTIME
        return PY_HEIDELTIME.extract_times(lang, text, publication_time)
    elif tool == 'tei2go':
        from text2story.annotators import TEI2GO
        return TEI2GO.extract_times(lang, text)

    raise InvalidTool


def extract_objectal_links(tool, lang, text):

    if tool in LOCAL_ANNOTATORS:
        local_annotator = importlib.import_module(tool)
        return local_annotator.extract_objectal_links(lang, text)

    if tool == "maverick":
        from text2story.annotators import MAVERICK
        return MAVERICK.extract_objectal_links(lang, text)


    raise InvalidTool

def extract_events(tool, lang, text):

    if tool in LOCAL_ANNOTATORS:
        local_annotator = importlib.import_module(tool)
        return local_annotator.extract_events(lang, text)

    if tool == 'custompt':
        from text2story.annotators import CUSTOMPT
        return CUSTOMPT.extract_events(lang, text)
    if tool == 'srl':
        from text2story.annotators import SRL
        return SRL.extract_events(lang, text)

    raise InvalidTool


def extract_semantic_role_links(tool, lang, text):

    if tool in LOCAL_ANNOTATORS:
        local_annotator = importlib.import_module(tool)
        return local_annotator.extract_semantic_role_links(lang, text)

    if tool == 'srl':
        from text2story.annotators import SRL
        return SRL.extract_semantic_role_links(lang, text)

    raise InvalidTool



import re
import string

import spacy
import transformers
import pandas as pd
import numpy as np

import nltk
from nltk.tokenize import sent_tokenize
from transformers import AutoModel, AutoTokenizer

nltk.download('punkt')

from itertools import zip_longest

from text2story.core.utils import bsearch_tuplelist
from transformers.pipelines import PIPELINE_REGISTRY

from text2story.annotators.SRL.srl_pipeline import SrlPipeline

from text2story.annotators.SRL.srl_model.model import SRLModel
from text2story.annotators.SRL.srl_model.config import SRLModelConfig

### Global Variables ###

"""SRL_TYPE_MAPPING = {
    #"TMP": "time",
    #"LOC": "location",
    #"ADV": "theme",  # Adverbial
    #"MNR": "manner",
    #"CAU": "cause",
    #"PRP":"cause",
    #"EXT": "theme",  # should be attribute -> changed for compatibility
    #"DIS": "theme",  # connection of two expressions -> should be theme -> changed for compatibility
    #"PNC": "purpose",
    #"PR": "purpose",
    #"NEG": "theme",  # should be attribute -> changed for compatibility | NEG = negation
    #"DIR": "path",  # should be setting -> changed for compatibility
    #"MOD": "instrument",  # MOD = Modal
    #"PRD": "theme",  # Secondary predicate -> should be attribute -> changed for compatibility
    #"ADJ": "theme",
    #"COM": "agent",  # COM = Comitative -> used to express accompaniment
    #"GOL": "goal",  # GOL = goal
    "REC": "instrument",  # REC = reciprocal
    "PAS": "passive", # passive verb
    "NSE": "agent",
    #"TML": "time",
    #"ASP": "theme"
}"""

SRL_TYPE_MAPPING = {
    # Time / Location / Path / Goal
    "TMP": "time",                 # temporal modifier
    "TML": "time",                 # temporal-locative label variants → still time in practice
    "LOC": "location",
    "DIR": "path",                 # direction/path; sometimes "goal" if 'to/into' marks endpoint
    "GOL": "goal",

    # Cause / Purpose / Reason
    "CAU": "cause",
    "PRP": "purpose",              # many tagsets use PRP for purpose; your draft put it under cause
    "PNC": "purpose",
    "PR":  "purpose",              # if your data uses PR for purpose, keep it

    # Manner / Means / Instrument / Medium
    "MNR": "manner",
    "MEANS": "means",              # (if present in your data)
    "INS": "instrument",           # (if present)
    "MED": "medium",               # (if present)

    # Extent / Amount / Distance / Attribute
    "EXT": "amount",               # degree/extent; if specifically distance, map to "distance"
    "NEG": "attribute",            # polarity is an attribute, not a role
    "PRD": "attribute",            # secondary predicate = depictives/resultatives → state/attribute
    "ADJ": "attribute",            # adjectival predication/modification → attribute
    "ASP": "attribute",            # aspectual operator (or create SRLINK_aspect if you prefer)
    # If you keep ADV as a bucket:
    "ADV": "manner",               # generic adverbials → prefer "manner" over "theme"

    # Participant relations
    "COM": "partner",              # comitative (with) → accompaniment
    "REC": "partner",              # reciprocal (each other) → symmetric co-participant
    "NSE": "agent",                # ⚠️ Ambiguous tag in some corpora; confirm your tagset definition
    "AGT": "agent",                # (if present)
    "PAT": "patient",              # (if present)
    "THM": "theme",                # (if present)

    # Voice / Modality / Discourse → not roles
    "PAS": "attribute",            # voice/diathesis → better as attribute (or SRLINK_voice)
    "MOD": "attribute",            # modality (can/could/might) → not instrument
    "DIS": "attribute",            # discourse connective → not a thematic role

    # Other plausible mappings in your inventory
    # "SRC": "source",             # if present
    # "GOL": "goal",               # already mapped above
    # "PRP": "purpose",            # already mapped above
}
pipeline = {}

def load(lang):
    """
    Used, at start, to load the pipeline for the supported languages.
    """
    nltk.download('punkt_tab')

    if lang == "pt":
        pipeline['srl_pt'] = transformers.pipeline(model="liaad/propbank_br_srl_albertina_100m_portuguese_ptpt_encoder",
                                               trust_remote_code=True)
        # regular expression to capture SRL tags
        pipeline["srl_re_tags"] = r".*[A][0-9]|AM"
        pipeline["ARGM_STR"] = "B-AM"
        # since allennlp does not have NER, download the spacy model
        if not (spacy.util.is_package('pt_core_news_lg')):
            spacy.cli.download('pt_core_news_lg')
        pipeline["pt"] = spacy.load('pt_core_news_lg')

    elif lang == "en":

        PIPELINE_REGISTRY.register_pipeline(
            "srl",
            pipeline_class=SrlPipeline,
            pt_model=SRLModel,
            type="text",
        )
        model = AutoModel.from_pretrained(
            "liaad/srl-en_roberta-large_hf", trust_remote_code=True
        )
        tokenizer = AutoTokenizer.from_pretrained(
            "liaad/srl-en_roberta-large_hf", trust_remote_code=True
        )
        #  Load the token classification pipeline with a pre-trained model
        pipeline['srl_en'] =  transformers.pipeline("srl", model=model, tokenizer=tokenizer, pipeline_class=SrlPipeline, lang="en")

        #pipeline["srl_re_tags"] = r"[BI]-(A\d|ARG\d|AM-[A-Z]+|ARGM-[A-Z]+)"
        #pipeline["srl_re_tags"] = r"[BI]-(C-)?(R-)?(A\d|ARG\d|AM-[A-Z]+|ARGM-[A-Z]+)"
        pipeline["srl_re_tags"] = r"[BI]-(?:C-)?(?:R-)?(?:A\d|ARG\d|AM-[A-Z]+|ARGM-[A-Z]+)"
        pipeline["ARGM_STR"] = "ARGM"

        # this coreference model is too heavy. Analyze another model to fit here
        #pipeline['coref_en'] =transformers.pipeline(model="VincentNLP/seq2seq-coref-t0-3b-partial-linear",
        #                                       trust_remote_code=True)
        pipeline['en'] = spacy.load('en_core_web_lg')


def process_srl_output(srl_output_list):
    """
    Aux function to process SRL english output

    @param srl_output:
    @return:
    """
    processed_output_list = []

    for srl_output in srl_output_list:
        for verb, (labels, tokens_with_labels) in srl_output.items():
            tokens = []
            predictions = []
            new_line_flag = False
            for token, label in tokens_with_labels:
                # Ignore <s> and </s> tokens
                if token not in ['<s>', '</s>']:
                    # Remove the leading extra token and add to new token
                    if token.startswith('Ġ') or new_line_flag:
                         cleaned_token = token.replace('Ġ','')
                         tokens.append(cleaned_token)
                         predictions.append(label)
                         new_line_flag = False
                    else:
                        if token == 'Ċ':
                            new_line_flag = True
                            continue
                        # this token is a continuation of the token before
                        tokens[-1] += token



            output_dict = {
                "tokens": tokens,
                "predictions": predictions
            }

            processed_output_list.append(output_dict)

    return processed_output_list


def _make_srl_df(lang, text):
    """
    Make a pandas DataFrame with the results from the SRL for each sentence in the text.
    Each row of the DataFrame is a frame from the SRL.

    @param text: The full text to annotate
    @return: List of pandas DataFrames with the contents of the SRL (values) for each frame (row) by word (column)
    """
    sentences = sent_tokenize(text)

    srl = []
    # se for um texto muito grande, da problema. O Allennlp tem um limite aqui
    for sent in sentences:

        toks = nltk.word_tokenize(sent)

        # allennlp has a limitation in the srl pipeline, they only accept sentences
        # with at most 512 tokens. So in here there is a test, and if necessary it splits
        # the sentence in batches of tokens
        if len(toks) > 256:

            result = []
            batch = 1
            len_batch = 256
            start_batch = 0
            end_batch = len_batch * batch

            while end_batch < len(toks):

                batch_sent = " ".join(toks[start_batch:end_batch])
                if abs(end_batch - start_batch) < 5:
                    batch = batch + 1
                    start_batch = end_batch
                    end_batch = len_batch * batch
                    continue


                if lang == "en":
                    result = pipeline["srl_en"].predict(sentence=batch_sent)
                elif lang == "pt":
                    try:
                         result = pipeline["srl_pt"].predict_json(batch_sent)
                    except RuntimeError:
                        print("Warning: Runtime Error -- %s" % batch_sent)
                        break
                    # there is a bug for some sentences, that there is no result from the srl
                    if isinstance(result, list):
                        continue

                batch = batch + 1
                start_batch = end_batch
                end_batch = len_batch * batch
                srl.append(result)
                for r in result:
                    print(len(r))
        else:
            if lang == "en":
                result = process_srl_output(pipeline["srl_en"](sent))

            elif lang == "pt":
                result = pipeline["srl_pt"](sent)
                #  there is a bug for some sentences, that there is no result from the srl?
            else:
                result = []
            srl.append(result)

    dfs_by_sent = []
    for sentence in srl:
        try:
            tags_by_frame = pd.DataFrame(columns=sentence[0]["tokens"])
        except IndexError as e:
            if len(sentence) == 0:
                continue
            #print("-->",sentence)
            #raise

        for i, frame in enumerate(sentence):
            #print("-->", i, len(sentence[i]["tokens"]),len(sentence[i]["predictions"]), sentence[i])
            #for i, frame in zip(np.arange(len(sentence["verb"])), sentence["verb"]):
            tags_by_frame.loc[i] = frame["predictions"]

        if tags_by_frame.shape[0] != 0:
            dfs_by_sent.append(tags_by_frame)

    return dfs_by_sent

def _normalize_sent_tags(sentence_df):
    """
    Normalize the frames retrieved from the SRL from one sentence.
    Each word must have only one label.

    @param sentence_df: DataFrame of the SRL each column is a word from the sentence and each row is the results of SRL
    for one frame.
    @return: List of the normalized tags
    @return: List of booleans of whether the tag is the beginning of the argument.
    """
    normalized_tags, begin_tags = [], []
    if sentence_df.size  == 0:
        return normalized_tags, begin_tags

    for col in np.arange(len(sentence_df.columns)):
        word_vals = sentence_df.iloc[:, col]

        word_vals = word_vals[word_vals != "O"]
        if word_vals.shape[0] == 1:
            normalized_tags.append(word_vals.iloc[0])
            begin_tags.append(word_vals.iloc[0].startswith("B"))
            continue
        verb_words = word_vals[word_vals.isin(["I-V", "B-V"])]
        if verb_words.shape[0] != 0:  # a) - verbo
            normalized_tags.append(verb_words.iloc[0])
            begin_tags.append(False)  # Event
            continue
        # b) - ARGM e ARG (o último e mais especifico tem prio)
        # TODO: em português as tags estao ligeiramente diferentes:
        # B-AM-TMP, B-A1, I-A2, etc
        arg_words = word_vals[word_vals.str.contains(pipeline["srl_re_tags"], regex=True)]
        if arg_words.shape[0] != 0:
            normalized_tags.append(arg_words.iloc[-1])  # desempate entre dois ARGM-X diferentes
            begin_tags.append(arg_words.iloc[-1].startswith("B"))
            continue
        else:
            print("\nNORMALIZATION ERROR - MULTIPLE TAG VALUES FOUND FOR WORD.")
            print(word_vals.values)
            print(sentence_df)
    #print(">>>",begin_tags)
    #print("-->",normalized_tags)
    return normalized_tags, begin_tags

def _find_events(normalized_tags, verb_tags, event_threshold=2):
    """
    Find words that belong to the same event.
    Each event can have a number of arguments between verbs and still be considered the same event.

    @param normalized_tags: result of normalized_sent_tags - a list of SRL tags for each word
    @param verb_tags: list of SRL tags for the algorithm to classify as event tags
    @param event_threshold: Threshold of non-verb arguments that can be found between verbs
    and still be considered part of the same event. n_args = event_threshold - 1
    @return: Boolean list of whether or not the word is part of an event
    """
    event_tags = []
    event_continue, event_begin = False, False
    for i, tag in zip_longest(np.arange(len(normalized_tags) - event_threshold), normalized_tags):
        if i is not None:
            if ("ARGM" in tag) & (normalized_tags[i + 1] in verb_tags):
                event_tags.append(True)
                event_begin = True
                continue

        if event_continue:
            event_tags.append(True)
        elif tag in verb_tags:
            event_tags.append(True)
            event_begin = True
        else:
            event_tags.append(False)

        if i is not None:
            conds = []
            for j in np.arange(1, event_threshold + 1):
                conds.append(normalized_tags[i + j] in verb_tags)

            if event_begin & any(conds):
                event_continue = True
            else:
                event_continue = False
                event_begin = False
        else:
            event_continue = False
            event_begin = False

    return event_tags

def _find_participants(begin_tags, event_tags):
    """
    Finds words that belong to the same participant or event and categorize them into participants.
    Each different argument represents a different participant. Arguments start at the begin tag (B).

    @param begin_tags: result of normalized_sent_tags - boolean list of tags that begin SRL arguments
    @param event_tags: result of find_events - boolean list of words that represent events
    @return: List of the participants that are represented by each word in the sentence
    """
    participant, i, participant_tags, event, is_event = False, 0, [], 1, False
    for btag, etag in zip(begin_tags, event_tags):
        if etag:
            is_event = True
        elif (not etag) & is_event:
            is_event = False
            event += 1

        if btag & (not etag):
            participant = True
            i += 1
            participant_tags.append("T" + str(i))
            continue

        if etag:
            participant_tags.append("EVENT" + str(event))
            participant = False
            continue

        if (not btag) & participant:
            participant_tags.append("T" + str(i))
        else:
            if len(participant_tags) > 0:
                participant_tags.append(participant_tags[-1])  # In case everything else fails, just join with previous participant
            else:
                # TODO: when participant_tags is empty what to do?
                i += 1
                participant_tags.append("T" + str(i))

    return participant_tags

def _srl_by_participant(srl_by_token, text, char_offset):
    """
    Organizes participants as words or expressions in the full text with their respective semantic role and character span.

    The semantic role is "EVENT" for event arguments and the SRL result for other participants.
    If the SRL result is a modifier, the most common in the participant is taken into account.

    The character span is a tuple - (start_char, end_char) -
    where the "start_char" is where the participant starts in the text and the "end_char" is where the participant ends.

    @param srl_by_token: DataFrame containing the results for the rest of the pipeline - namely, the participant references
    @param text: The full text to be annotated
    @param char_offset: The current char position in the text. Each word in the text increments it
    @return: Dict list with the participant, its semantic role and the its character position span in the text

    @note: CAN MALFUNCTION IF SRL DOES NOT FIND FRAMES IN EVERY SENTENCE.
    The char positions are checked in the entire text, if there is a word before in a sentence not recognized by the SRL
    it will malfunction slightly (give wrong position values).
    """
    result_list = []
    ARGM_STR = pipeline["ARGM_STR"]
    for participant in srl_by_token["participant"].unique():
        rows = srl_by_token[srl_by_token["participant"] == participant]
        tags = rows["tag"]
        if participant.startswith("EVENT"):
            sem_role_type = "EVENT"
        elif any(ARGM_STR in tag for tag in tags):
            argm_tags = [tag for tag in tags if ARGM_STR in tag]
            #print("-->",argm_tags[0].split("-")[-1])
            sem_role_type = SRL_TYPE_MAPPING[argm_tags[0].split("-")[-1]]
        else:
            sem_role_type = "PARTICIPANT"

        char_spans = []
        for token in rows.index:
            if all(c in string.punctuation or c.isspace() for c in token):
                # Pure punctuation - literal search
                pos = text.find(token, char_offset)
                if pos != -1:
                    char_spans.append((pos, pos + len(token)))
                    char_offset = pos + len(token)
                else:
                    char_spans.append(-1)
            else:
                # Strip all punctuation from both ends to get the word part
                word_part = token.strip(string.punctuation)

                if word_part:
                    pattern = r'\b' + re.escape(word_part) + r'\b'
                    match = re.search(pattern, text[char_offset:])

                    if match:
                        start_match = char_offset + match.start()
                        end_match = char_offset + match.end()

                        char_spans.append((start_match, end_match))
                        #char_offset += len(word_part) # raises error if there is punctuation in the end
                        #char_offset += len(token) # raises eror if there is punctuation in the start of token
                        char_offset = end_match
                    else:
                        char_spans.append(-1)
                else:
                    char_spans.append(-1)
                #print("-->", char_spans, "--", token)

        result_list.append({
            "participant": ' '.join(rows.index), "sem_role_type": sem_role_type,
            "char_span": (char_spans[0][0], char_spans[-1][1]) # the start of the first token in the span, and the end of the last token in the span
        })

    return result_list, char_offset

def _srl_pipeline(sentence_df, text, char_offset, verb_tags, event_threshold=3):
    """
    Pipeline to retrieve participants and events by their order in the text, with their semantic roles and character spans.
    Each iteration of the pipeline takes a dataframe of the SRL results for a sentence -> as returned by _make_srl_df.

    @param sentence_df: The DataFrame with the SRL results with the structure shown in _make_srl_df
    @param text: The full text to be annotated
    @param char_offset: The current character position offset to start looking for character spans (recursive)
    @param verb_tags: SRL tags that represent verbs -> defaults to ["B-V", "I-V"]
    @param event_threshold: Threshold of non-verb arguments that can be found between verbs
    and still be considered part of the same event. number_of_args = event_threshold - 1

    @return: df_by_participant -> Pandas DataFrame with each participant, their semantic roles and character spans
    character_offset -> The current character position offset in the full text
    """
    # 2. Remove words out of vocabulary in every frame #
    oov_mask = []
    for name, values in sentence_df.items():
        oov_mask.append(any(~(values == "O")))
    sentence_df = sentence_df.loc[:, oov_mask]
    sentence_df = sentence_df.apply(lambda x: x.sort_values(ascending=False).values)  # Begin tags in the end always

    # 3. Normalize SRL tags for each word in the sentence dataframe #
    normalized_tags, begin_tags = _normalize_sent_tags(sentence_df)

    # 4. Define events #
    event_tags = _find_events(normalized_tags, verb_tags=verb_tags, event_threshold=event_threshold)

    #print("---", event_tags)
    #print()
    # 5. Find and categorize expressions into participants #
    participant_tags = _find_participants(begin_tags, event_tags)

    # Putting together a dataframe for the characteristics of each word in this sentence
    if sentence_df.size == 0:
        return pd.DataFrame(), 0
    df = pd.DataFrame({
        "tag": normalized_tags, "is_begin_tag": begin_tags, "is_event": event_tags, "participant": participant_tags
    }, index=sentence_df.columns)

    # 6. Find semantic roles and character spans for each participant in the sentence #
    df_by_participant, character_offset = _srl_by_participant(df, text, char_offset)

    return df_by_participant, character_offset


def classify_semantic(tok):
    if tok.ent_type_ != '':
        return tok.ent_type_.lower().capitalize()
    if tok.pos_ == 'PRON':
        return 'Pronoun'
    elif tok.pos_ == 'DET':
        return 'Determiner'
    elif tok.pos_ == 'NOUN':
        return 'NominalPhrase'
    elif tok.pos_ == 'VERB':
        if 'Infinitive' in tok.tag_ or tok.tag_ in ['VB', 'VINF']:
            return 'PurposeClause'
        return 'Clause'
    elif tok.pos_ in ['ADP', 'PART']:
        next_token = tok.nbor(1) if tok.i + 1 < len(tok.doc) else None
        if next_token and next_token.pos_ == 'VERB':
            if 'Infinitive' in next_token.tag_ or next_token.tag_ in ['VB', 'VINF']:
                return 'PurposeClause'
    if hasattr(tok, 'dep_') and tok.dep_ in ['xcomp', 'advcl', 'obl:purpose']:
        return 'PurposeClause'
    return 'Arg'


def get_participant_tags(text, participant_lst, lang):
    """
    it adds POS tags and NER tags to the participant list

    @param text: the original text to be processed as narrative
    @param list[tuple[int,int]]: A list of character span of the participant
    candidates
    Returns
    ------
    list[tuple[tuple[int, int], str, str]]
       a list of participants of narratives, where the first is the character
       span of that participant, the second elements is the POS tag, anf third
       element is the NER tag, if it is a entity
    """
    participant_lst = sorted(participant_lst, key=lambda x:x[0])
    tagged_participants = {}

    doc = pipeline[lang](text)
    tagged_participants_lst = []

    for idx, offset_participant in enumerate(participant_lst):
        span = doc.char_span(offset_participant[0], offset_participant[1])
        if span:
            head = span.root
            semantic_class = classify_semantic(head)
            tagged_participants[idx] = (offset_participant, head.pos_, semantic_class)

        # Tokens fora dos spans
    for tok in doc:
        pos = bsearch_tuplelist(tok.idx, participant_lst)
        if pos == -1:
            tagged_participants_lst.append(((tok.idx, tok.idx + len(tok.text)), tok.pos_, "UNDEF"))

        # Adiciona os participantes encontrados
    for idx in range(len(participant_lst)):
        if idx in tagged_participants:
            tagged_participants_lst.append(tagged_participants[idx])

    tagged_participants_lst.sort(key=lambda x: x[0][0])
    return tagged_participants_lst



def extract_participants(lang, text):
    """
    Main function that applies the SRL pipeline to extract participant entities from each sentence.

    @param lang: The language of the text
    @param text: The full text to be annotated
    Returns
    -------
    list[tuple[tuple[int, int], str, str]]
        the list of participants identified where each participant is represented by a tuple
    """
    # 1. DATAFRAME WITH THE SRL RESULTS OF EVERY FRAME FOR EACH SENTENCE IN THE TEXT #
    dfs_by_sent = _make_srl_df(lang, text)

    # FIND EVENTS - PIPELINE #
    character_offset, srl_participants_list = 0, []
    for sent_df in dfs_by_sent:
        df_by_participant, character_offset = _srl_pipeline(sent_df, text, character_offset, verb_tags=["B-V", "I-V"],
                                                      event_threshold=2)

        srl_participants_list.append(df_by_participant)


    srl_participants_list = [item for sublist in srl_participants_list for item in sublist]  # flatten list
    srl_df = pd.DataFrame(srl_participants_list)
    if len(srl_df) == 0:
        return []

    participants_df = srl_df[srl_df["sem_role_type"] == "PARTICIPANT"]
    participants_lst = participants_df["char_span"].values.tolist()

    # TODO: these participants still dont have POS and NER tags
    participants_tagged_lst = get_participant_tags(text, participants_lst, lang)
    #participants_lst = [ (p, "NOUN", "Per") for p in participants_lst]

    return participants_tagged_lst

def extract_events(lang, text):
    """
    Main function that applies the SRL pipeline to extract event entities from each sentence.
    Joins every event participant from each sentence in the text.

    @param lang: The language of the text
    @param text: The full text to be annotated

    @return: Pandas DataFrame with every event entity and their character span
    """
    # 1. DATAFRAME WITH THE SRL RESULTS OF EVERY FRAME FOR EACH SENTENCE IN THE TEXT #
    dfs_by_sent = _make_srl_df(lang, text)

    # FIND EVENTS - PIPELINE #
    character_offset, srl_participants_list = 0, []
    for sent_df in dfs_by_sent:
        df_by_participant, character_offset = _srl_pipeline(sent_df, text, character_offset, verb_tags=["B-V", "I-V"],
                                                      event_threshold=3)
        srl_participants_list.append(df_by_participant)

    srl_participants_list = [item for sublist in srl_participants_list for item in sublist]  # flatten list
    srl_df = pd.DataFrame(srl_participants_list)
    if len(srl_df) == 0:
        return []
    return srl_df[srl_df["sem_role_type"] == "EVENT"]

def extract_semantic_role_links(lang, text):
    dfs_by_sent = _make_srl_df(lang, text)

    character_offset, srl_by_sentence = 0, []
    for sent_df in dfs_by_sent:
        df_by_participant, character_offset = _srl_pipeline(sent_df, text, character_offset, verb_tags=["B-V", "I-V"],
                                                      event_threshold=3)
        srl_by_sentence.append(pd.DataFrame(df_by_participant))
    return srl_by_sentence
